(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     66796,       1699]
NotebookOptionsPosition[     26566,        790]
NotebookOutlinePosition[     63829,       1602]
CellTagsIndexPosition[     63786,       1599]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["FindMinumum and FindRoot for 1D functions.", "Title"],

Cell["\<\
 Code Generation for FindMinimum[ ] and FindRoot[ ] 
of a 1D function using MathCode\
\>", "Subtitle"],

Cell["\<\
Copyright \[Copyright] MathCore AB 1998-2006\
\>", "Copyright"],

Cell[TextData[{
 "The function ",
 StyleBox["FindMinimum[] ",
  FontWeight->"Bold"],
 " finds a local minumum of given function and given starting value. Function \
cannot be translated by MathCode directly because\n--- its argument is a \
symbolic expression \n --- numerical algorithms for minimization are not \
included in MathCode runtime library\n The solution is to write a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " function which replicates ",
 StyleBox["FindMinimum[ ]",
  FontWeight->"Bold"],
 " operation using compilable subset.\n This notebook illustrates how to do \
that for finction with one argument. \n ",
 StyleBox["FindRoot[ ]",
  FontWeight->"Bold"],
 " can be implemented using ",
 StyleBox["FindMinumum[ ] ",
  FontWeight->"Bold"],
 "as well."
}], "Text"],

Cell["\<\

This is how FindMinumum works in Mathematica\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindMinimum", "[", 
  RowBox[{
   RowBox[{"Cos", "[", "x", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "1"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "1.`"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", "\[Rule]", "3.1415926555856286`"}], "}"}]}], "}"}]], "Output",\

 CellChangeTimes->{3.5112584528501863`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.51125798262183*^9, 3.511258453240806*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetDirectory", "[", 
  RowBox[{"ToFileName", "[", 
   RowBox[{"{", 
    RowBox[{"$MCRoot", ",", "\"\<Demos\>\"", ",", "\"\<FindMinimum\>\""}], 
    "}"}], "]"}], "]"}]], "Input"],

Cell[BoxData["\<\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\FindMinimum\"\>"], \
"Output",
 CellChangeTimes->{3.5112584532564306`*^9}]
}, Open  ]],

Cell[TextData[{
 "\nThis function ",
 StyleBox["LookupMin[ ]",
  FontWeight->"Bold"],
 " replicates ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " ",
 StyleBox["FindMinimum[ ]",
  FontWeight->"Bold"],
 ". The arguments are: \n",
 StyleBox["iters",
  FontWeight->"Bold"],
 " - maximal number of iterations\n",
 StyleBox["k",
  FontWeight->"Bold"],
 " - trial step multiplier\n",
 StyleBox["alphastart",
  FontWeight->"Bold"],
 " - a small number needed for computation of approximation of derivatives\n\
",
 StyleBox["startp",
  FontWeight->"Bold"],
 " - start value for search\n",
 StyleBox["accuracy",
  FontWeight->"Bold"],
 " - algorithm stoops when steps are less than accuracy\n",
 StyleBox["debug",
  FontWeight->"Bold"],
 " - whether additional debugging is needed.\nThe minimized function name is \
not passed to ",
 StyleBox["LookupMin[ ]",
  FontWeight->"Bold"],
 ". The name is always ",
 StyleBox["Func[ ]. ",
  FontWeight->"Bold"],
 "Later we will see how to adjust this name \"on the flight\"."
}], "Text"],

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"LookupMin", "[", 
     RowBox[{
      RowBox[{"Integer", " ", "iters_"}], ",", 
      RowBox[{"Real", " ", "k_"}], ",", 
      RowBox[{"Real", " ", "alphastart_"}], ",", 
      RowBox[{"Real", " ", "startp_"}], ",", 
      RowBox[{"Real", " ", "accuracy_"}], ",", 
      RowBox[{"Integer", " ", "debug_"}]}], "]"}], "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"Real", ",", "Real"}], "}"}]}], ":=", "\[IndentingNewLine]", 
   RowBox[{"Module", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{
        RowBox[{"Real", " ", "p"}], "=", "startp"}], ",", 
       RowBox[{
        RowBox[{"Integer", " ", "i"}], "=", "0"}], ",", 
       RowBox[{
        RowBox[{"Real", " ", "currentAccuracy"}], "=", "99999."}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"Real", " ", "alpha"}], "=", "alphastart"}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"Real", " ", "df"}]}], "}"}], ",", "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{"While", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{"i", "<", "iters"}], ")"}], "&&", 
         RowBox[{"(", 
          RowBox[{"currentAccuracy", ">", "accuracy"}], ")"}]}], ",", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"df", "=", 
          RowBox[{
           RowBox[{"(", 
            RowBox[{
             RowBox[{"Func", "[", 
              RowBox[{"p", "+", "alpha"}], "]"}], "-", 
             RowBox[{"Func", "[", "p", "]"}]}], ")"}], "/", "alpha"}]}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"p", "=", 
          RowBox[{"p", "-", 
           RowBox[{"k", "*", "df"}]}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"alpha", "=", 
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"k", "*", "df"}], "]"}], "/", "2"}]}], ";", 
         "\[IndentingNewLine]", 
         RowBox[{"i", "=", 
          RowBox[{"i", "+", "1"}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"currentAccuracy", "=", 
          RowBox[{"Abs", "[", 
           RowBox[{"k", "*", "df"}], "]"}]}], ";", "\[IndentingNewLine]", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"debug", ">", "0"}], ",", 
           RowBox[{"Print", "[", 
            RowBox[{"{", 
             RowBox[{
              RowBox[{"Func", "[", "p", "]"}], ",", "p", ",", 
              "currentAccuracy"}], "}"}], "]"}]}], "]"}], ";"}]}], 
       "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
      RowBox[{"{", 
       RowBox[{"p", ",", 
        RowBox[{"Func", "[", "p", "]"}]}], "}"}]}]}], "\[IndentingNewLine]", 
    "\[IndentingNewLine]", "]"}]}]}]], "Input"],

Cell["Functions to minimize", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Func1", "[", 
     RowBox[{"Real", " ", "u_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{
    RowBox[{"Cos", "[", "u", "]"}], "+", "4"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Func2", "[", 
     RowBox[{"Real", " ", "u_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{"u", " ", 
    RowBox[{"Sin", "[", 
     RowBox[{"u", "+", "1"}], "]"}]}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"Func2", "[", "u", "]"}], ",", 
   RowBox[{"{", 
    RowBox[{"u", ",", 
     RowBox[{"-", "2"}], ",", "2"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJwtmXk4Vd/3xxEqmTI1IHOjDBFSrFLSRFFEZUiRkMqnEJJoQBGSpCQZEiVN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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{{-2, 2}, {-0.2401252439984678, 1.6829418127116607`}},
  PlotRangeClipping->True,
  PlotRangePadding->{
    Scaled[0.02], 
    Scaled[0.02]}]], "Output",
 CellChangeTimes->{3.5112584533345547`*^9}]
}, Open  ]],

Cell["\<\

Try to minimize in Mathematica\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"mm", "=", 
  RowBox[{"FindMinimum", "[", 
   RowBox[{
    RowBox[{"Func2", "[", "v", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"v", ",", 
      RowBox[{"-", "1"}]}], "}"}]}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "0.24012524415530795`"}], ",", 
   RowBox[{"{", 
    RowBox[{"v", "\[Rule]", 
     RowBox[{"-", "0.520268992107989`"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5112584533501797`*^9}]
}, Open  ]],

Cell[TextData[{
 "Try to minimize using our ",
 StyleBox["LookupMin",
  FontWeight->"Bold"],
 " function"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Func", "[", "u_", "]"}], ":=", 
   RowBox[{"Func2", "[", "u", "]"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"lm", "=", 
  RowBox[{"LookupMin", "[", 
   RowBox[{"500", ",", "0.1", ",", "0.01", ",", 
    RowBox[{"-", "1."}], ",", "0.0000001", ",", "0"}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "0.5202694059459493`"}], ",", 
   RowBox[{"-", "0.24012524415513598`"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5112584533658047`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Print", "[", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{
      RowBox[{"mm", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
      "-", 
      RowBox[{"lm", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}], 
     ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{"v", "/.", " ", 
        RowBox[{
        "mm", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}], ")"}],
       " ", "-", 
      RowBox[{
      "lm", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}]}], "}"}],
    "]"}], "\[IndentingNewLine]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "1.7197354651443675`*^-13"}], ",", 
   "4.1383796023364283`*^-7"}], "}"}]], "Print",
 CellChangeTimes->{3.511258453381429*^9}]
}, Open  ]],

Cell[TextData[{
 "\nCompile ",
 StyleBox["LookupMin",
  FontWeight->"Bold"],
 " to C++. Note that we use Func2 instead of Func replacing it on the \
flight."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"BuildCode", "[", 
  RowBox[{"MacroRules", "\[Rule]", 
   RowBox[{"{", 
    RowBox[{"Func", "\[Rule]", "Func2"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "Func", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Func}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5112584536783004`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "Func", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Func}],
  Editable->False]], "Print",
 CellChangeTimes->{3.51125845370955*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "Func", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Func}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5112584537408*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 3 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5112584558813972`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], "\[IndentingNewLine]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5112584595376005`*^9}],

Cell[BoxData[
 RowBox[{"LinkObject", "[", 
  RowBox[{"\<\"\\\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\FindMinimum\\\\\
Globalml.exe\\\"\"\>", ",", "77", ",", "8"}], "]"}]], "Output",
 CellChangeTimes->{3.5112584595376005`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"cm", "=", 
  RowBox[{"LookupMin", "[", 
   RowBox[{"500", ",", "0.1", ",", "0.01", ",", 
    RowBox[{"-", "1."}], ",", "0.0000001", ",", "0"}], "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "0.5202694059459493`"}], ",", 
   RowBox[{"-", "0.24012524415513598`"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5112584595532255`*^9}]
}, Open  ]],

Cell["How accurate are C++ computations ?", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Print", "[", 
  RowBox[{"{", 
   RowBox[{
    RowBox[{
     RowBox[{"mm", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}], 
     "-", 
     RowBox[{"cm", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}], 
    ",", 
    RowBox[{
     RowBox[{"(", 
      RowBox[{"v", "/.", " ", 
       RowBox[{"mm", "\[LeftDoubleBracket]", "2", "\[RightDoubleBracket]"}]}],
       ")"}], " ", "-", 
     RowBox[{"cm", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}]}],
    "}"}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"-", "1.7197354651443675`*^-13"}], ",", 
   "4.1383796023364283`*^-7"}], "}"}]], "Print",
 CellChangeTimes->{3.51125845956885*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell["\<\

Using FindRoot[ ]\
\>", "Subtitle"],

Cell[TextData[{
 "\nFinding the root in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ". This function is normally used for solving non-linear equations."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"mathroot", "=", 
  RowBox[{"x", "/.", 
   RowBox[{"FindRoot", "[", 
    RowBox[{
     RowBox[{
      RowBox[{"Sin", "[", "x", "]"}], "\[Equal]", 
      RowBox[{"Log", "[", "x", "]"}]}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", "0.5"}], "}"}]}], "]"}]}]}]], "Input"],

Cell[BoxData["2.219107148913746`"], "Output",
 CellChangeTimes->{3.5112584595844746`*^9}]
}, Open  ]],

Cell["\<\

The problem is reduced to finding minumum of the residual function.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FindMinimum", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Sin", "[", "x", "]"}], "-", 
      RowBox[{"Log", "[", "x", "]"}]}], ")"}], "^", "2"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", "0.5"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"2.294993478814369`*^-19", ",", 
   RowBox[{"{", 
    RowBox[{"x", "\[Rule]", "2.2191071493680594`"}], "}"}]}], "}"}]], "Output",\

 CellChangeTimes->{3.5112584596157246`*^9}]
}, Open  ]],

Cell["\<\

Prepare the resifual function for MathCode compilation\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"Func3", "[", 
     RowBox[{"Real", " ", "u_"}], "]"}], "\[Rule]", "Real"}], ":=", 
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"Sin", "[", "u", "]"}], "-", 
      RowBox[{"Log", "[", "u", "]"}]}], ")"}], "^", "2"}]}], ";"}]], "Input"],

Cell["\<\

Compute the result\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"BuildCode", "[", 
   RowBox[{"MacroRules", "\[Rule]", 
    RowBox[{"{", 
     RowBox[{"Func", "\[Rule]", "Func3"}], "}"}]}], "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "Func", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Func}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5112584599438453`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "Func", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Func}],
  Editable->False]], "Print",
 CellChangeTimes->{3.51125845999072*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "Func", "}"}]}],
  SequenceForm[
  "Functions without types are not compiled :", {$CellContext`Func}],
  Editable->False]], "Print",
 CellChangeTimes->{3.5112584600219693`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 4 function(s)\"\>"], "Print",
 CellChangeTimes->{3.511258462162567*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5112584657718954`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"croot", "=", 
  RowBox[{
   RowBox[{"LookupMin", "[", 
    RowBox[{
    "500", ",", "0.1", ",", "0.01", ",", "0.5", ",", "0.0000001", ",", "0"}], 
    "]"}], "[", 
   RowBox[{"[", "1", "]"}], "]"}]}]], "Input"],

Cell[BoxData["2.2191067840591425`"], "Output",
 CellChangeTimes->{3.5112584657875204`*^9}]
}, Open  ]],

Cell["Check accuracy", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"croot", "-", "mathroot"}]], "Input"],

Cell[BoxData[
 RowBox[{"-", "3.648546034185074`*^-7"}]], "Output",
 CellChangeTimes->{3.5112584658031454`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell["CleanMathCodeFiles[Confirm -> False, CleanAllBut -> {}];", "Input"]
},
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{678, 905},
WindowMargins->{{133, Automatic}, {8, Automatic}},
ShowCellTags->True,
FrontEndVersion->"8.0 for Microsoft Windows (32-bit) (November 7, 2010)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "\nModify the definitions below to change the default appearance of all \
cells in \na given style.  Make modifications to any definition using \
commands in the \nFormat menu.\n", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "\nThe options defined for the style below will be used at the \
Notebook level.\n", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "\nThe cells in this section define styles used for input and output \
to the \nkernel.  Be careful when modifying, renaming, or removing these \
styles, \nbecause the front end associates special meanings with these style \
names.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "\nThe following styles are useful for numbered equations, figures, \
etc.  They \nautomatically give the cell a FrameLabel containing a reference \
to a \nparticular counter, and also increment that counter.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6], 
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "\nThe cells below define styles useful for making hypertext \
ButtonBoxes.  The \n\"Hyperlink\" style is for links within the same \
Notebook, or between \nNotebooks.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, 
             FontVariations -> {"Underline" -> True}, FontColor -> 
             RGBColor[0, 0, 1], Background -> None, 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
         "\nThe following styles are for linking automatically to the on-line \
help \nsystem.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Open]]}, 
     Open]]}, Visible -> False, FrontEndVersion -> 
  "8.0 for Microsoft Windows (32-bit) (November 7, 2010)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[557, 20, 59, 0, 55, "Title"],
Cell[619, 22, 112, 3, 61, "Subtitle"],
Cell[734, 27, 73, 2, 54, "Copyright"],
Cell[810, 31, 791, 21, 193, "Text"],
Cell[1604, 54, 69, 3, 55, "Text"],
Cell[CellGroupData[{
Cell[1698, 61, 167, 5, 48, "Input"],
Cell[1868, 68, 213, 7, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2118, 80, 74, 1, 48, "Input"],
Cell[2195, 83, 469, 10, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2701, 98, 202, 5, 48, "Input"],
Cell[2906, 105, 133, 2, 47, "Output"]
}, Open  ]],
Cell[3054, 110, 1028, 37, 283, "Text"],
Cell[4085, 149, 2720, 72, 369, "Input"],
Cell[6808, 223, 37, 0, 25, "Text"],
Cell[6848, 225, 221, 7, 48, "Input"],
Cell[7072, 234, 247, 8, 48, "Input"],
Cell[CellGroupData[{
Cell[7344, 246, 193, 6, 48, "Input"],
Cell[7540, 254, 9489, 161, 255, "Output"]
}, Open  ]],
Cell[17044, 418, 55, 3, 55, "Text"],
Cell[CellGroupData[{
Cell[17124, 425, 219, 7, 48, "Input"],
Cell[17346, 434, 248, 7, 47, "Output"]
}, Open  ]],
Cell[17609, 444, 117, 5, 25, "Text"],
Cell[17729, 451, 134, 4, 48, "Input"],
Cell[CellGroupData[{
Cell[17888, 459, 186, 4, 48, "Input"],
Cell[18077, 465, 192, 5, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[18306, 475, 591, 18, 69, "Input"],
Cell[18900, 495, 182, 5, 26, "Print"]
}, Open  ]],
Cell[19097, 503, 169, 6, 55, "Text"],
Cell[CellGroupData[{
Cell[19291, 513, 162, 4, 48, "Input"],
Cell[19456, 519, 316, 8, 23, "Print"],
Cell[19775, 529, 313, 8, 23, "Print"],
Cell[20091, 539, 312, 8, 23, "Print"],
Cell[20406, 549, 122, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[20565, 555, 95, 2, 69, "Input"],
Cell[20663, 559, 228, 5, 23, "Print"],
Cell[20894, 566, 227, 4, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[21158, 575, 186, 4, 48, "Input"],
Cell[21347, 581, 192, 5, 47, "Output"]
}, Open  ]],
Cell[21554, 589, 51, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[21630, 593, 527, 15, 48, "Input"],
Cell[22160, 610, 181, 5, 26, "Print"]
}, Open  ]],
Cell[22356, 618, 79, 2, 48, "Input"],
Cell[22438, 622, 46, 3, 61, "Subtitle"],
Cell[22487, 627, 173, 5, 55, "Text"],
Cell[CellGroupData[{
Cell[22685, 636, 297, 9, 48, "Input"],
Cell[22985, 647, 89, 1, 47, "Output"]
}, Open  ]],
Cell[23089, 651, 92, 3, 55, "Text"],
Cell[CellGroupData[{
Cell[23206, 658, 281, 9, 48, "Input"],
Cell[23490, 669, 214, 6, 50, "Output"]
}, Open  ]],
Cell[23719, 678, 79, 3, 55, "Text"],
Cell[23801, 683, 308, 10, 48, "Input"],
Cell[24112, 695, 43, 3, 55, "Text"],
Cell[CellGroupData[{
Cell[24180, 702, 183, 5, 48, "Input"],
Cell[24366, 709, 316, 8, 23, "Print"],
Cell[24685, 719, 313, 8, 23, "Print"],
Cell[25001, 729, 316, 8, 23, "Print"],
Cell[25320, 739, 120, 1, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25477, 745, 77, 2, 48, "Input"],
Cell[25557, 749, 228, 5, 23, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[25822, 759, 234, 7, 48, "Input"],
Cell[26059, 768, 90, 1, 47, "Output"]
}, Open  ]],
Cell[26164, 772, 30, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[26219, 776, 60, 1, 48, "Input"],
Cell[26282, 779, 110, 2, 47, "Output"]
}, Open  ]],
Cell[26407, 784, 79, 2, 48, "Input"],
Cell[26489, 788, 73, 0, 48, "Input"]
}
]
*)

(* End of internal cache information *)
