(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    185793,       3794]
NotebookOptionsPosition[    180475,       3619]
NotebookOutlinePosition[    180867,       3636]
CellTagsIndexPosition[    180824,       3633]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Interpolation in 1D", "Title"],

Cell["Example - how to save interpolation table in C++ file.", "Subtitle"],

Cell["", "Title"],

Cell["\<\

\
\>", "Author"],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " has built-in InterpolatingFunction object which can be used as \
approximations of arbitrarly complerx functions. This notebook explains using \
interpolatingFunction with one floating-point argument. The original and \
helping functions as well as a table with constants is converted to C++ using \
MathCode. We later verify the accuracy of the results, i.e. compare original \
function, interpolated function and MathCode variant of the interpolated \
function. .  \n"
}], "Text"],

Cell["", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086603665657425`*^9, 3.508660544735337*^9, 
  3.5086607510764456`*^9, 3.5086607988414593`*^9, 3.508660870824913*^9}]
}, Open  ]],

Cell["\<\
SetDirectory[$MCRoot<>$PathnameSeparator<>\"Demos\"<>$PathnameSeparator<>\"\
Interpolation-1D\"];\
\>", "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["The interpolated function", "Section"],

Cell["\<\
This function will be appoximate using InterpolatedFuncion object.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"f", "[", "x_", "]"}], ":=", 
  RowBox[{
   RowBox[{"Sin", "[", "x", "]"}], "+", 
   SuperscriptBox["x", "2"], "-", 
   RowBox[{"2", "x"}]}]}]], "Input"],

Cell[TextData[{
 "We create ",
 StyleBox["InterpolatedFuncion",
  FontWeight->"Bold"],
 " object",
 StyleBox[" g ",
  FontWeight->"Bold"],
 "from function",
 StyleBox[" f ",
  FontWeight->"Bold"],
 "on an interval",
 StyleBox[" [-2,3] ",
  FontWeight->"Bold"],
 "using 11 points. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"g", "=", 
  RowBox[{"FunctionInterpolation", "[", 
   RowBox[{
    RowBox[{"f", "[", "x", "]"}], ",", 
    RowBox[{"{", 
     RowBox[{"x", ",", " ", 
      RowBox[{"-", "2."}], ",", "3."}], "}"}], ",", 
    RowBox[{"InterpolationPoints", "\[Rule]", "11"}]}], "]"}]}]], "Input",
 CellChangeTimes->{3.5086610399008737`*^9}],

Cell[BoxData[
 TagBox[
  RowBox[{"InterpolatingFunction", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "2.`"}], ",", "3.`"}], "}"}], "}"}], ",", "\<\"<>\"\>"}], 
   "]"}],
  False,
  Editable->False]], "Output",
 CellChangeTimes->{3.50866036675324*^9, 3.5086605448447104`*^9, 
  3.5086607511701946`*^9, 3.5086607989352083`*^9, 3.508660870903037*^9}]
}, Open  ]],

Cell[TextData[{
 "Now, we use \"hidden\" feaures of  ",
 Cell[BoxData[
  StyleBox["InterpolatingFunction",
   FontWeight->"Bold"]]],
 " object and extract the tables of length ",
 StyleBox["nx1",
  FontWeight->"Bold"],
 " with values for  variables",
 StyleBox[" ",
  FontWeight->"Bold"],
 Cell[BoxData[
  FormBox["x1", TraditionalForm]],
  FontWeight->"Bold"],
 "and",
 StyleBox[" ",
  FontWeight->"Bold"],
 Cell[BoxData["y"],
  FontWeight->"Bold"],
 "."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"x1", "=", 
   RowBox[{"g", "\[LeftDoubleBracket]", 
    RowBox[{"3", ",", "1"}], "\[RightDoubleBracket]"}]}], ";", 
  RowBox[{"nx1", "=", 
   RowBox[{"Length", "[", "x1", "]"}]}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{
  RowBox[{"If", " ", "[", 
   RowBox[{
    RowBox[{"$VersionNumber", " ", "<", "6"}], ",", "\[IndentingNewLine]", 
    RowBox[{"y", "=", 
     RowBox[{"g", "\[LeftDoubleBracket]", 
      RowBox[{"4", ",", "2"}], "\[RightDoubleBracket]"}]}], ",", " ", 
    "\[IndentingNewLine]", 
    RowBox[{"If", " ", "[", 
     RowBox[{
      RowBox[{"$VersionNumber", ">", "7"}], ",", "\[IndentingNewLine]", " ", 
      RowBox[{"y", "=", 
       RowBox[{"Flatten", "[", 
        RowBox[{"g", "[", 
         RowBox[{"[", "4", "]"}], "]"}], "]"}]}], ",", "\[IndentingNewLine]", 
      
      RowBox[{"y", "=", 
       RowBox[{"g", "\[LeftDoubleBracket]", 
        RowBox[{"4", ",", "3"}], "\[RightDoubleBracket]"}]}]}], 
     "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", "]"}], 
  ";"}]], "Input",
 CellChangeTimes->{{3.5086606593432446`*^9, 3.5086607304360847`*^9}, {
  3.5086607674199867`*^9, 3.5086607764667454`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", "y"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
  "7.090702573174318`", ",", "6.438211785191379`", ",", "5.817695631982345`", 
   ",", "5.229458398114625`", ",", "4.673689698348237`", ",", 
   "4.15046242204372`", ",", "3.6597319429768445`", ",", "3.201336603126446`",
    ",", "2.7749994651906897`", ",", "2.380331314832929`", ",", 
   "2.0168348801101734`", ",", "1.6839102223417899`", ",", 
   "1.380861239976666`", ",", "1.106903214951137`", ",", 
   "0.8611713197284716`", ",", "0.6427299927960622`", ",", 
   "0.4505830809816258`", ",", "0.28368463863670784`", ",", 
   "0.14095026661477258`", ",", "0.0212688681193019`", ",", 
   RowBox[{"-", "0.07648530602191173`"}], ",", 
   RowBox[{"-", "0.15344045323772992`"}], ",", 
   RowBox[{"-", "0.21071498855336057`"}], ",", 
   RowBox[{"-", "0.24940541997997812`"}], ",", 
   RowBox[{"-", "0.270574461395797`"}], ",", 
   RowBox[{"-", "0.2749084736255334`"}], ",", 
   RowBox[{"-", "0.27523951178550565`"}], ",", 
   RowBox[{"-", "0.2716858604242085`"}], ",", 
   RowBox[{"-", "0.2643616257333199`"}], ",", 
   RowBox[{"-", "0.25337642608480904`"}], ",", 
   RowBox[{"-", "0.23883509473928766`"}], ",", 
   RowBox[{"-", "0.2208373955318681`"}], ",", 
   RowBox[{"-", "0.19947775230661358`"}], ",", 
   RowBox[{"-", "0.1748449928333633`"}], ",", 
   RowBox[{"-", "0.14702210790145265`"}], ",", 
   RowBox[{"-", "0.11608602624367403`"}], ",", 
   RowBox[{"-", "0.08210740590090504`"}], ",", 
   RowBox[{"-", "0.04515044259322876`"}], ",", 
   RowBox[{"-", "0.005272695617261736`"}], ",", "0.03747506825813973`", ",", 
   "0.08304901247442364`", ",", "0.13141234957074088`", ",", 
   "0.18253544095376117`", ",", "0.23639587727933176`", ",", 
   "0.2929785392279691`", ",", "0.35227563850937527`", ",", 
   "0.4142867389847429`", ",", "0.47901875784959635`", ",", 
   "0.5464859468739363`", ",", "0.6167098537505806`", ",", 
   "0.6897192636565137`", ",", "0.765550121185754`", ",", 
   "0.844245432865435`", ",", "0.9258551505195092`", ",", 
   "1.0104360357963778`", ",", "1.188771463237071`", ",", 
   "1.3798357076052792`", ",", "1.5843100319532715`", ",", 
   "1.8029655844723624`", ",", "2.036655038562893`", ",", 
   "2.286303349098633`", ",", "2.5528977250959644`", ",", 
   "2.837476927492947`", ",", "3.1411200080598665`"}], "}"}]], "Output",
 CellChangeTimes->{3.5086603667844896`*^9, 3.5086607514201913`*^9, 
  3.5086607989664574`*^9, 3.508660870934286*^9}]
}, Open  ]],

Cell["\<\
Now, global variables and their initial values are declared for generating \
using MathCode. \
\>", "Text"],

Cell[BoxData[
 RowBox[{"With", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"xx1", "=", "x1"}], ",", 
     RowBox[{"yy", "=", "y"}], ",", 
     RowBox[{"nn1", "=", "nx1"}]}], "}"}], ",", "\[IndentingNewLine]", 
   RowBox[{"Declare", "[", "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{
      RowBox[{
       RowBox[{"Real", "[", "nn1", "]"}], " ", "x1a"}], "=", "xx1"}], ",", 
     "\[IndentingNewLine]", 
     RowBox[{
      RowBox[{
       RowBox[{"Real", "[", "nn1", "]"}], " ", "ya"}], "=", "yy"}]}], 
    "\[IndentingNewLine]", "]"}]}], "\[IndentingNewLine]", "]"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implementation  of interpolationfunctions", "Section"],

Cell["Function 1", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"PolyInt", "[", 
   RowBox[{"xa_", ",", "ya_", ",", "x_"}], "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"n", ",", "i", ",", "m", ",", 
      RowBox[{"ns", "=", "1"}], ",", "den", ",", "dif", ",", "dift", ",", 
      "ho", ",", "hp", ",", "w", ",", "c", ",", "d", ",", "y"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"n", "=", 
      RowBox[{"Length", "[", "xa", "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"dif", "=", 
      RowBox[{"Abs", "[", 
       RowBox[{"x", "-", 
        RowBox[{
        "xa", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}], 
       "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"c", "=", 
      RowBox[{"Range", "[", "n", "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"d", "=", 
      RowBox[{"Range", "[", "n", "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"Do", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"dift", "=", 
         RowBox[{"Abs", "[", 
          RowBox[{"x", "-", 
           RowBox[{
           "xa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], 
          "]"}]}], ";", "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"dift", "<", "dif"}], ",", "\[IndentingNewLine]", 
          RowBox[{
           RowBox[{"ns", "=", "i"}], ";", 
           RowBox[{"dif", "=", "dift"}]}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"c", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
         "=", 
         RowBox[{
         "ya", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{
         RowBox[{"d", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
         "=", 
         RowBox[{
         "ya", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}], 
       ",", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"i", ",", "n"}], "}"}]}], "\[IndentingNewLine]", "]"}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"y", "=", 
      RowBox[{"ya", "\[LeftDoubleBracket]", "ns", "\[RightDoubleBracket]"}]}],
      ";", "\[IndentingNewLine]", 
     RowBox[{"ns", "--"}], ";", "\[IndentingNewLine]", 
     RowBox[{"Do", "[", "\[IndentingNewLine]", 
      RowBox[{
       RowBox[{
        RowBox[{"Do", "[", "\[IndentingNewLine]", 
         RowBox[{
          RowBox[{
           RowBox[{"ho", "=", 
            RowBox[{
             RowBox[{
             "xa", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], 
             "-", "x"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"hp", "=", 
            RowBox[{
             RowBox[{"xa", "\[LeftDoubleBracket]", 
              RowBox[{"i", "+", "m"}], "\[RightDoubleBracket]"}], "-", 
             "x"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"w", "=", 
            RowBox[{
             RowBox[{"c", "\[LeftDoubleBracket]", 
              RowBox[{"i", "+", "1"}], "\[RightDoubleBracket]"}], "-", 
             RowBox[{
             "d", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}]}]}], 
           ";", "\[IndentingNewLine]", 
           RowBox[{"den", "=", 
            RowBox[{"ho", "-", "hp"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{"den", "=", 
            RowBox[{"w", "/", "den"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
            "d", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "=", 
            
            RowBox[{"hp", "*", "den"}]}], ";", "\[IndentingNewLine]", 
           RowBox[{
            RowBox[{
            "c", "\[LeftDoubleBracket]", "i", "\[RightDoubleBracket]"}], "=", 
            
            RowBox[{"ho", "*", "den"}]}]}], ",", "\[IndentingNewLine]", 
          RowBox[{"{", 
           RowBox[{"i", ",", 
            RowBox[{"n", "-", "m"}]}], "}"}]}], "]"}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"y", "+=", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"2", "*", "ns"}], "<", 
            RowBox[{"n", "-", "m"}]}], ",", 
           RowBox[{"c", "\[LeftDoubleBracket]", 
            RowBox[{"ns", "+", "1"}], "\[RightDoubleBracket]"}], ",", 
           RowBox[{
            RowBox[{"ns", "--"}], ";", 
            RowBox[{"d", "\[LeftDoubleBracket]", 
             RowBox[{"ns", "+", "1"}], "\[RightDoubleBracket]"}]}]}], 
          "]"}]}]}], ",", "\[IndentingNewLine]", 
       RowBox[{"{", 
        RowBox[{"m", ",", 
         RowBox[{"n", "-", "1"}]}], "}"}]}], "]"}], ";", 
     "\[IndentingNewLine]", "y"}]}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"General", "::", "spell1"}], "MessageName"], 
  RowBox[{
  ":", " "}], "\<\"New symbol name \\\"\[NoBreak]\\!\\(dift\\)\[NoBreak]\\\" \
is similar to existing symbol \\\"\[NoBreak]\\!\\(dif\\)\[NoBreak]\\\" and \
may be misspelled. \\!\\(\\*ButtonBox[\\\"\[RightSkeleton]\\\", ButtonStyle->\
\\\"Link\\\", ButtonFrame->None, \
ButtonData:>\\\"paclet:ref/message/General/spell1\\\", ButtonNote -> \
\\\"General::spell1\\\"]\\)\"\>"}]], "Message", "MSG",
 CellChangeTimes->{3.508660367018862*^9, 3.508660751467066*^9, 
  3.5086607991539555`*^9, 3.508660871121784*^9}]
}, Open  ]],

Cell["\<\
Funktion 2 locates actual interval which is created by  dividing of the total \
interval into pieces.\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Locate", "[", 
   RowBox[{"xx_", ",", "x_"}], "]"}], ":=", "\[IndentingNewLine]", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"n", ",", "ju", ",", "jm", ",", "jl", ",", "j"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"n", "=", 
      RowBox[{"Length", "[", "xx", "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"jl", "=", "0"}], ";", "\[IndentingNewLine]", 
     RowBox[{"ju", "=", 
      RowBox[{"n", "+", "1"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"While", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"ju", "-", "jl"}], ">", "1"}], ",", "\[IndentingNewLine]", 
       RowBox[{
        RowBox[{"jm", "=", 
         RowBox[{"Floor", "[", 
          FractionBox[
           RowBox[{"ju", "+", "jl"}], "2."], "]"}]}], ";", 
        "\[IndentingNewLine]", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"(", 
            RowBox[{"x", "\[GreaterEqual]", 
             RowBox[{"xx", "[", 
              RowBox[{"[", "jm", "]"}], "]"}]}], ")"}], "\[Equal]", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{
             "xx", "\[LeftDoubleBracket]", "n", "\[RightDoubleBracket]"}], 
             ">", 
             RowBox[{
             "xx", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}], 
            ")"}]}], ",", "\[IndentingNewLine]", 
          RowBox[{"jl", "=", "jm"}], ",", "\[IndentingNewLine]", 
          RowBox[{"ju", "=", "jm"}]}], "\[IndentingNewLine]", "]"}]}]}], 
      "\[IndentingNewLine]", "]"}], ";", "\[IndentingNewLine]", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"x", "\[Equal]", 
        RowBox[{
        "xx", "\[LeftDoubleBracket]", "1", "\[RightDoubleBracket]"}]}], ",", 
       "\[IndentingNewLine]", 
       RowBox[{"j", "=", "1"}], ",", "\[IndentingNewLine]", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{"x", "\[Equal]", 
          RowBox[{
          "xx", "\[LeftDoubleBracket]", "n", "\[RightDoubleBracket]"}]}], ",",
          "\[IndentingNewLine]", 
         RowBox[{"j", "=", 
          RowBox[{"n", "-", "1"}]}], ",", "\[IndentingNewLine]", 
         RowBox[{"j", "=", "jl"}]}], "\[IndentingNewLine]", "]"}]}], 
      "\[IndentingNewLine]", "]"}]}]}], "\[IndentingNewLine]", 
   "]"}]}]], "Input"],

Cell["The main function calls two above functions.", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"InterpolatingFunction1D", "[", "x_", "]"}], ":=", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", 
     RowBox[{"n", ",", "j", ",", "m", ",", "k"}], "}"}], ",", 
    "\[IndentingNewLine]", 
    RowBox[{
     RowBox[{"n", "=", 
      RowBox[{"Length", "[", "x1a", "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"j", "=", 
      RowBox[{"Locate", "[", 
       RowBox[{"x1a", ",", "x"}], "]"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"m", "=", "4"}], ";", "\[IndentingNewLine]", 
     RowBox[{"k", "=", 
      RowBox[{"Min", "[", 
       RowBox[{
        RowBox[{"Max", "[", 
         RowBox[{
          RowBox[{"j", "-", 
           RowBox[{"Floor", "[", 
            FractionBox[
             RowBox[{"m", "-", "1."}], "2"], "]"}]}], ",", "1"}], "]"}], ",", 
        
        RowBox[{"n", "+", "1", "-", "m"}]}], "]"}]}], ";", 
     "\[IndentingNewLine]", 
     RowBox[{"PolyInt", "[", 
      RowBox[{
       RowBox[{"x1a", "\[LeftDoubleBracket]", 
        RowBox[{"k", "|", 
         RowBox[{"k", "+", "m", "-", "1"}]}], "\[RightDoubleBracket]"}], ",", 
       
       RowBox[{"ya", "\[LeftDoubleBracket]", 
        RowBox[{"k", "|", 
         RowBox[{"k", "+", "m", "-", "1"}]}], "\[RightDoubleBracket]"}], ",", 
       "x"}], "]"}]}]}], "\[IndentingNewLine]", "]"}]}]], "Input"],

Cell["\<\
Deklarations of three above functions are necessary in order to use them with \
MathCode\
\>", "Text"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"PolyInt", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Real", "[", "nn_", "]"}], "  ", "xa_"}], ",", 
      RowBox[{
       RowBox[{"Real", "[", "nn", "]"}], "  ", "ya_"}], ",", 
      RowBox[{"Real", " ", "x_"}]}], "]"}], "\[Rule]", "Real"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "Integer", ",", "Integer", ",", "Integer", ",", "Integer", ",", 
     "\[IndentingNewLine]", "Real", ",", "Real", ",", "Real", ",", "Real", 
     ",", "Real", ",", "Real", ",", 
     RowBox[{"Real", "[", "nn", "]"}], ",", 
     RowBox[{"Real", "[", "nn", "]"}], ",", "Real"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"Locate", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Real", "[", "_", "]"}], "  ", "xx_"}], ",", 
      RowBox[{"Real", " ", "x_"}]}], "]"}], "\[Rule]", "Integer"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "Integer", ",", "Integer", ",", "Integer", ",", "Integer", ",", 
     "Integer"}], "}"}]}], "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"Declare", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"InterpolatingFunction1D", "[", 
     RowBox[{"Real", " ", "x_"}], "]"}], "\[Rule]", "Real"}], ",", 
   RowBox[{"{", 
    RowBox[{"Integer", ",", "Integer", ",", "Integer", ",", "Integer"}], 
    "}"}]}], "]"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Code generation", "Section"],

Cell[TextData[{
 StyleBox["CompilePackage",
  FontFamily->"Courier New"],
 " generates C++ source code."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CompilePackage", "[", "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "f", "}"}]}],
  SequenceForm["Functions without types are not compiled :", {$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660367331358*^9, 3.5086607517795615`*^9, 
  3.5086607994820757`*^9, 3.50866087143428*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "f", "}"}]}],
  SequenceForm["Functions without types are not compiled :", {$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660367331358*^9, 3.5086607517795615`*^9, 
  3.5086607994820757`*^9, 3.5086608714811544`*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "f", "}"}]}],
  SequenceForm["Functions without types are not compiled :", {$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660367331358*^9, 3.5086607517795615`*^9, 
  3.5086607994820757`*^9, 3.508660871512404*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 3 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508660367331358*^9, 3.5086607517795615`*^9, 
  3.5086607994820757`*^9, 3.5086608736842513`*^9}]
}, Open  ]],

Cell[TextData[{
 StyleBox["MakeBinary",
  FontFamily->"Courier New"],
 StyleBox[" translates C++ code into runnable executable.",
  FontFamily->"Times New Roman"]
}], "Text"],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", "]"}]], "Input"],

Cell[CellGroupData[{

Cell["Test", "Subsection"],

Cell[TextData[{
 "We test automatically generated function and compare it with ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "'s own  ",
 StyleBox["InterpolatingFunction",
  FontFamily->"Courier New"],
 " object"
}], "Text"],

Cell[" Evaluate the original function", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"f", "[", 
  RowBox[{"-", "1.1"}], "]"}]], "Input"],

Cell[BoxData["2.5187926399385647`"], "Output",
 CellChangeTimes->{3.5086603731906576`*^9, 3.5086607572794914`*^9, 
  3.5086608052476273`*^9, 3.50866087726233*^9}]
}, Open  ]],

Cell[TextData[{
 "Evaluation of  ",
 StyleBox["Mathematica's approximation ",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"g", "[", 
  RowBox[{"-", "1.1"}], "]"}]], "Input"],

Cell[BoxData["2.51879489779062`"], "Output",
 CellChangeTimes->{3.508660373221907*^9, 3.5086607573107405`*^9, 
  3.508660805278877*^9, 3.508660877309205*^9}]
}, Open  ]],

Cell["\<\
Evaluation of our implemented function that approximates the original \
function. It will be later translated to C++ .\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InterpolatingFunction1D", "[", 
  RowBox[{"-", "1.1"}], "]"}]], "Input"],

Cell[BoxData["2.51879489779062`"], "Output",
 CellChangeTimes->{3.5086603733625307`*^9, 3.5086607573576155`*^9, 
  3.5086608053257513`*^9, 3.508660877356079*^9}]
}, Open  ]],

Cell[TextData[{
 "We ",
 StyleBox["install",
  FontSlant->"Italic"],
 " the generated C++ code."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660373471904*^9, 3.5086607574357395`*^9, 
  3.5086608054038754`*^9, 3.5086608774810777`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InterpolatingFunction1D", "[", 
  RowBox[{"-", "1.1"}], "]"}]], "Input"],

Cell[BoxData["2.51879489779062`"], "Output",
 CellChangeTimes->{3.5086608278098383`*^9, 3.508660877496702*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell[TextData[{
 "Find quality of our approximation (we run in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " kernel this)",
 StyleBox[".",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"f", "[", "x", "]"}], "-", 
     RowBox[{"InterpolatingFunction1D", "[", "x", "]"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "2."}], ",", "3"}], "}"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJws2Wk4ld/XAGBFZKhUlESESkI0SGI7SaKIhPxKSJKQJPOQMVNmhRDHPM/z
tBzTIUlCkpBEUYoGUaHX3v/3U93XWnuttZ/zjJedJre0r61mYGAQZGNgwP/+
6Yof//fvX8PKfxloh0Ghc178U+VvYse1v+PQY0GYufWDmBKLcpC1mubcni/Y
tACXXaWIYjv65+0HbE9e5t4qtCn+NkPsKImvExID9L55NbPWIDHL3ehG5Me9
ayPtGbZgnJIAHemjii1ObSQ+Xy3ajvZeP8Un1Yht/FOTowP9CRvYOVmDPeq9
7ulT1Fl1Yw+1jPR7tnDtGXr87o+4fgGJH+Ye6ULWbMEHOLNIvWxp3m5EOcgv
256MLeUyW9aNNhoUKHjGY8+GtCi+QGO+iidkH2JzrhKrfYHK8rtVZ0PJfEHx
e3rQvf4rZ7MCsLudI+/1IL1/388be2NTavxe9iBRUd//eNywi55/3dqLfmtx
G3Xbk/grRY1e1OGcYRpwi8zr/8W+FyWkHLGg3CDz7reI7EXWT9tvLZiQ+ap2
p/YixZ//2RcZkP7JkVm9iJP/s4u5HqnHyLcSHzvp5imoha0Vtj+qF5VZr/Mb
UMOmhsg696J7MYn3w0+Q+AhVtxfp0fZHqCpghy90i/Yi0SlaNMMRcnw+7f/R
g/5s1E6okiL9LgqX96Cncu+TbcRIvveGWz0o4apdpqgI+T1kLuzsQdbBa/JH
+Un+yKXOF4hSHl0Su5UcPyd3mxdo48ieKq2NpH4s7/qV47n/dDONieTnyB3p
Rn76b9qdlpdx3OCj33N0wcuqS2oBm6FB9GEX+tMT+pr6GVswq+dhJ6I49M54
9pD80INy7WhTkumcbCfJb0rb3Ibet839mW0l+WHrJluR3zYe5ivVJH9B2LcZ
ddYZ8B2nYnv65R2oRxcYP6iusiYuZahKQbYZOusOXCf1rFJn41CwWvMLE2Pi
VwJh4agpjHqxRZtYNlfdGiT5Lln5HyHr8/Ol00Gt4YlUlRSJ39FtygZTE9m5
yb3EGp8PFcCjLO67p/mIv4BBKZSd8VVy5camXJc1LYeur9+Z89aT+P6k/yqB
6XB32LpVJP7Om70WBAaQDvqzhOPTNfQ6kHPN57n1A5vmr3cTQGcH33DSNLF6
23ID3GoMSu6ewKbcPFFPgyDT39dWvcU2tmC3aYQ0FnOxAwPYgnaeW5tgUONk
aVQHWc/76EQzzM2WOrY0Y1NdB9uagfOBkPxcHfaoDROlBfYdiWDYXUHqfZ7J
bQGVwX8teoWk3pgzeytccbcO9M8i/XV8DFvBTXBYoyqZzK/6La0VopvPbJqK
I+s780Zaodispn/bA7L+dwo7HTpZ98afDiHrZ1r20eFDXoyRqx+Z5zWLIh1W
aTGL5HmQ+BfTk3Tg+2E3OeT0v/29RnQ4Ev0+b50tqZdgtrJe+6j2bWSJ7fl6
40r9m0O0w7dMSdxvfLgVAjz2/0m6TOptnE5thVShROjWI/0Wjl5uBWjl8Fml
RfZPf8/aCgPmrqcOqBFzMea0wA/2T+xXlUi9oHKFFlhfqN8ddYzMM8zX2gyi
2m0PWg6R+n2GlGYwjE3j372HeMuOzU3gfGzzmJ4gOT6Kfy0a4cGIV4b/NlK/
zbaSBk9EjPZPsZP5P8kZAoy3PfuxbQ35vdhSyutg2UK+6vTyIq4Xr7hQA4eK
tx3PmyU2cT5eCZo6AWuGp7BpmRfUysFi/teTde+xGdxmUSlQFfq0b70k9jCa
zofaUaWt1C5sT4FS22zo9yl+091GfCNwMA3YO0JND9SQ/LblnGhIqzcp6C8l
1v/IcB/ki2UWXPKJOW6X2kBfGpuSQCaxsq6lPVjFjtxvphKzyReGAFNwycvr
ccTe78Ri4bGHnwDHA9LPrDmLCl1mkqU6/sRP/f2ywezi6qUFT5L/ZUdXPvzT
6Fd57EI8MrBYBFKH7w5O3CT7VXj+twwiGbtWOeoQXw2VqQOx+eQz28+S4yN0
5XU9NH+yf9hwivh01o0GMBhRe3v1OLGF4x0a/HzBv3ftMdJ/yUS0EYJbv9nm
HcKmPjbpbQSR6tY6LUkS33fSrgl0qTe1Hu3EHkXL0c3wNep4nMJ2Epf25mkB
P3/u8XdcxF9OhLXADtcpCb/1xAxCf1ugwrreUWwtmXf78qVWOGsS0di1isS5
6otb4aPuNfY7f//i+mMqi63gqXZUd+scNiXN7xgdeBTWJdV+xfZ8ZHGLDkVS
7yaNJrEZRMdj6KAmUn6AaYysf/2xjA5jWwPdst6QfMObdDq4sF+mq78k8cBb
nXTY9E+K81sXNk13eiWe+53p4sN2kv9uvJwOyh8GUo82kfjty4/oMPQ678tw
7f/7Nh3snnke8S7HptrOKNBhXaOO1+5CbEHxnf9aIaNM9GlHFpnPjbG8FVDW
ItetFJJ/PceoFfrjuw03J5B5eCX+tYB1WFpW5UNs46n0qBZg9nH6fimM1NM4
wN8CSQ7q8gyBJH/ye3wzyFoI+qV5k3n92Dmbofvyz+eqbuR48aQ6NwHDyYSr
EbfI+uw/Eo0QK2uTf/gGWa/L40wDaXHl+dcmZD/po3YNYLJ5OkhIj7jw4vE6
+MNM66NrkvlzNwXXQNSfqB2WaiTOu721ClreHSsplSf9ADGVw+6i+6+Vhck8
V+VCs8HzgtJGVn5iXWXFdBhcXlB9toVYpvwVFQ5lFHpGcJL6Cn3FsRCqYVal
y0ast39jGEz+5JvdxkQ87jDoBEoJvXtGlv5gh2eKWkDCiSCjlHniRC2HQPj1
iRJj9o3Y5AprNGhFzneJfSZebhVPgpyjBcwz48QVRzemgWHAdgfHgT/k+D2p
ywP6rl86AQ3YtJb4qxUg+CwvWL2a+FsDqgIXu6stnKUkf86HtQYkm7sPPsog
9TreWdZDgIW/5WUqNsXq3uoGGNuIUnfGYRsrW5+gQbRx7qacEGzBMrfcRphl
MTlt7U/if/9qNcHpQh7vA16k/tOQ6SZI03te/cuF1C9kudsMy0v3vtXYEe9U
WtMC+unyez2syfpgHu8WKFH/YXzCnPTXuPGzBTh+ZseymJB4IbdBK5jFG3c/
vUTiM6uqW4GmtHVtuC6pJ8e98vzg/fRMUUcTm5p/VHvl/I7wdeRRI/vXNwil
Q5fsscIhJWK5mw10EB399oEqT+ZVNpqgg7d/1o5rMqTfBVGGNhiSNNLbK0Xq
qzRuaAOZfu7QL3tJvG8XVxuEu3e2FguTepU669rgk4jPkj0/qRet+Xfl+us8
elhuK5ln7fYROiTembVa5iTHLxZWrs8F3sy0JjZSr0HZe+V52HR5yI+J7O9z
zQk65N3g4jqz/BvHXaT/tALzxqdnNixgj9bVpreCcZWXT+83bIa5WyqtUGMk
WxvzmVjv+lALcLHMfL80gU1hhusr12dBupjgW2xP/qjJZmjXNTAZH/hNrp+1
Rs0gtLQpLquH5MefeNoE/Wc8WaXpxE559xpB6ofM8bkGbMHNB/tW3l/ivjhV
V5P1XlLbaICmLk4ezyf9zzevr4d0v8Nt2g+I65fmykE160l9WAip12IxVArT
Ty6XdfqRuOua4mI4sM4vWcWZOCdZMwf6JXljfWyJU2Rj0sFZqyCUZknmvxTd
SAU+W6V7S6YkzrBEjQFaVL+rnCGxCtfW+3C13MLW8QKx1IdDFsD8atm8TIu4
aNDIDXIWIo2+qRH/e3Q2AjR49+hJniA+fUEnHmaP1apbypN+GlWRKfDgsuaJ
rMMkritilAmDVEcpIVES79Q/WwR3m9j3GO0kjj9xrwR2jifxJ/CS/br25JSB
+Z52ti3riFM+QRWwqxmsOs9M8q035NVAgcXsfNi/BXI/33CvDubyeSZYv2N7
npdeaIDY53lvVD5jG++MbaeB/DdKj8/4Atn/WGAjvN30sp02jD1aVKnYBN6H
bjQs9WNTykammmC33lK5XDfJpwoGNMMTx4g8xyek/o4zvC1g9WhXalkTqT+7
n9oCnLXVj77Vknmm0ra1QtmQRrhkOVk/5+PXCheW3/lZFpA4Z+6HVvgj4OCe
lUn6ay0eXTnfj7PZTVBJ/shlTzocv5poIRRH8ueqq+kw7nvgilEUiX9aXrn+
AjLoFxKCyfqdAmvaYF/7xbOv72ELbtu4tQ26pr4qb/Eg8xo9394Gtuw+x847
Ecfpb24DbomtB8JvE3sXLtGh6myu6DMLsp+JwUE6GNgoCrCZkn7pEzl0+BfR
y33qMpmncuQmHVJKr3P46pF+e14L00Hl5d/VjZrY1E3Tna0w9Svs95Iqce6B
G60QzCMyK6dE6h3tXGgBKbmqD47HSH+uIdcW6L2kPlx2iNTLC/zeDA7uo73f
JIi9GC43A2+SXYfkHtJf1Ka2CYzHEiqztpF50zh0G4GJSbpgYhP5/Rjzw2mQ
uas1TYiD9Iu66N0AX82/RCQsz2NPNTHWQUSQV8DreWzPee1dK++bedweW74R
G/04UAWuMwpW4e+xaUYbuMsg4cqRScowid9x5iyBul4p02/9xC9HfxfCcqWw
gXYHqf9I/EImCIjxD6xuIY6WG0kBSsIWndJ6Yu17jxLgynrO7quVxAGOqlHg
7cmqzlVMXJcW6wOp31e3t+QQewbzikCL6eIJ+zTisek+b5jon2vYlUg8/K8v
EpjVZo71xxBLxnMlwJ7ayUq/CGJ3kagUUJUYO3jkPvHxCMZMCNr4cl/sXTK/
b8LDQsj16cpUdSL7nVMdLIbOuTbh37exKW4Tg6WwfrBm+8VrZH3xjvxK2K9e
FsNmRPL5Yt2rQQvyN9fqk/z+aflaiEyhsvOpk7jlLj+AUq44/86TxKn3WWjQ
5xfF6K5I+st6NtFgbiHYQ+IoWZ9ee6cRtlj6/R0+QLyfY3sTHBn2cAwVxzaW
PFHVBPqazj/QbuzZdYfPNINzo+2tGQESV6/ubYa4g1afk7Zhj7rknGuB2vRr
17U2Y9sofGhpgaGtRu8Z1mEXndaQaIWlQH2jYmbscMvmoFbYsXjuzRUG0t91
/8r3laL1mQub/vzC5+N1VxE6GI8q9zb9wLaRfGBMBy9tpHnnC7bW2I2Ileuj
5chT4Y/YnukzFXRolpE+1TeKzZC3/cXK9Zwl1uw7iM157ONbOqzZLqJ4uI/k
Z599T4fdIfy1E8+www3PvaHDqX9bjkS3Efd8a6OD+W3OUpVGbKnQw9kr94P3
rPvna7CLxHZ50CFblzE3s4zUO9epRoeOtsXd+gWkX91RVjp8OvorZW0WNvWg
R30rsOfN7KhOxjaOKDZrBfEdU3E34km/+kmmVtAIH9vC+5Dsv0gxduX5uHoo
siMUe/ZWn2ALFH3oCtrnTeL1aFMzvNBvZx5yJfOlGLk2wfeORu9gexJn1B1s
hEOFZS5fzMk89+xdaKC7s+DXYxPs0aG1Tg3gEJVpe9aAHK8nhpP1UOUYZ1Go
iU05ci+yBoRMb6cPqGLTfkRnVEGIluroKiWy37UzGRVgsveXrs4h0s9J+VYJ
dHI/C3eXIPuda5AtApnVaU8zdpP1gotjucD+5tzx3zxkfexh8xRwaBN1E9pE
zPF5OQ5GS/9VnGEnZvRcCIfT1P5vdkxkvozVTK5QFpwvnrg0hz1e42wFO5x9
r7f9IpY/qn0fAq9dSpmdmSPP42NRMWCIWHlOjBEvBJumQ7vYqLbVG2zKdC9z
DhzYWhnysI/Ec9YEFQDLrCnjJB1b8PWF6FKwHTqGNtJIvk1QWjkMtW9ylqsm
+VoC1EooSqZ9Dc7FpkKVQS3whsbsrUjDpj104KuHey7Wpm8fE6uwtgHMmp1M
WhtD5hX7JkGDi+f5BqXDSbxD4QsNWhR/cF0KxJ5N2JfcCJLiHZq+3tjhYx5q
TcC4xqm1346sT73n0AzW384yMFiTee70/2mGgeFdx/Zex+6+0mnbAkodi/ba
xmQ/HsdHWiCvorfI9T/sUSYuhVbYkprzOU2b1NPZHdYKXmFeu7vOkP3OGfa1
wrSr/pV5ZTJfTBYHHfTM9ycIImybjd+P0IGmw/xK7QipbyipRwex48Mb70hh
ayXpXKfDA4ky9YS95PhsN7Zceb5tu+/fKkTmRVomdLBgNmn6up04RlSdDn3f
ZZe2cpN6g7OidEBvN8geX088XLbQCllPP9hasJD6cb61rbCpqj4/ioHsV8Lu
diu4pz2YrPv9Ex+/l/G8rTAZbin84Tu24PetlS2g7a5kuGEa25iR9VQL1N3Y
9kh2Ans0J6qjeeV9ZLb3ygg2JWf0eDP8lUxUK+sm9c47czSB2XZ73+EnZD3/
PuNG6GZRb2BuxuZ0s8yiQdro70P/lZP1jML7AdY/677lXYDtabHoXgfO1Zk5
uZlkvcSGqho4G6kruPyI5JtcmK0AlwMb1E5FkXntjH6XQUZP++3wYGwGuYnZ
EljeJN8s5EHyq+8Y5INY6dxnSydiD5bPWaB3vpCr/DZxk7dBGuRHCZmdMiX9
Df2GH8LAwaHQ8Muk3rHRMwHA1Pew8rUesazCBhOQstMcFdIitvu06i4YcLGy
WqmR9Q9d2CMgoKxJulyJ2J5iHgdlOm4Xl4+R/JOFT5KB4+FMbrgk6c/QlJoD
soez+17vIfmlwhMFYPrSZEloJ9l/Y4VPMdRx958t34xNHQ+xK4fJ8jDHZQ5S
L/azdiVw6alRTzGTeie5BKrBMrru2+uFH9iONnfrIFrGgVf4O7Zx+31OgKb+
/SesPhNnHw9deR9xmLIsHyf5N4JdacC7NfXB8jA2laa5qxFUKg3qT73CptUe
bmkE2wtbPoR3Ywsuceg3QeL88/WDT0i9yczRJuiICTwi3Iwdnjy08r7068gJ
Y6s67G6Oe89XvlcGFgPKy0n9Z3cPt8BZp4ri5QJsG5XcsBZw4bEZPJWFPRoz
M9wCGVV7GSOSSX/qIcFW6NF/v28wDttz0UyvFZYXEnSEH5B6jI6erbD3kZ67
VQhxh35iK+gc5cwo98PW2slU1Aqer590LXuQ+pF25a2Q5+wzf8oZu8g0u6AV
BrYpCEbYEj9Mj28FpppfqoOWpN7ILbdWkLpYdFv4Gtlv0aZzrWDw50aclSE5
HocSVt63A+KEm8svkP2c3tHfAmVyw5+XtbBngwr8WmB0MJpL9TSpf8Js38r3
squWQsQJbCknk5ZmkN3OZjYoT/yu+VwzmNY2hwrLYDNMF/Q2Qfgl90qr/dic
NWdON8Fk/OzafzvJ8UjT2NIIXPI50qrbyfH24raiAWXo6sUILhI/mW298j3C
9ypXmIXUS9QwrIOWuvA+Kwayn5N2K9fbrMHppfLf3/Hxjgj6VwWnHtefVZ3G
HqUM6JXDbMjQItMHbMqBDUGlEHf3b07jW2zPyI6CYpg2OspyrJd4njcxF6K1
9MvnO0l+nm9MJiged7xaRsdm8MpzSoVIofIGiRqSX/H060OQ39x381Mpif/J
MQyCCcYf2zPzidtTRW9D6M+NHVczif3k99qC7ISUk2AycebTo0Ew9lJz93Ac
qcf9KvAh3Kdb9z16QOJPbboew0hmntTmALK/qp2fMiAg9unIcy+SL3uiPQek
Az8FB7uS+F+lkALwtRCdWnOL7F+u6H0JSF46FdNkTuolSjwug1dnzE56mJB8
qflTFSAmkZa8oIttfCHgVjX08TdrlmuS9foWn2vAff3Y0m01bMGrz/TqoHtG
4OJneZKf7v8HwGUUrc2Swe4O2LTyfS784nKFqRRZ/+tjLg0cSuI3jQiT/MPs
FY0gmFpDi+Mn/X9slGiCJ1GvrS9sJfnKEXFNYOu7wMe1kfim1XIT8NlvfdrN
RvpbBeo3A/2ajHMIE9nPz+6sZrilp7vn9PI3fP7dlpxpBp5Tdi+ZF7Bn14Sv
nK9NR6J8mr9hU4I+XF75/hQtkfb8jE29v2vl/Obe9uKt/ASx+om0FgDW2ZDf
I9iCLbJVLWD+Z718xQCpl7CqqQU2fZb4ZNuDrXU4oXHl/fiNeuz+TmzPRNaV
55Fpp6XKdCs2Q4JGSgusrw/6mdVA5ku19mmBqvzslGvVpN+g/cUWuJLYriVU
SvJ5b+1pAbawj8sjedicU+afmqHMgzk/PoPU97BLbQZDm12X9KlknuZ07WZg
uaLMyh2HPXpm80ITFJ27Wvkiiuz3Un9UE1xU8r4WGkLqv2fb3QRMB5M3n/En
+V7PCxohX5jWyOKFXeR9UrIRGNYs83vZEUsPbKRBzhxfp4I1yU8KnAY4/+GY
y5/rxJfXpNVDRptz/51LpD7NcLQGHh1wOJGtQ+Z9feVkNQQ/ti0a0SBxyXcx
lWB7xzJIjYJtfC2TpQyujVyfv3uUHG8Osd0loK9malp2gOTbUg8VgYLAZSSw
i5j1imoWSAX9l6uz4xs5fzzXpYHQnC5P0FYS/9FkkwTcxtr3GjiJK+QeRsPa
p2e//2QlznDceh/+Hj5jJMZIvGPA8DZ8pZ7qNFqcxVZ+8McW3rErH304R6xe
wHwf+hwoGR1fiQej+aOh7Z38ZoZJbE/RfUGJUKN+1PPwO2LGRZNUoAoduEjt
xab8ydfMhagQybaXncRN6dsLwW9B7BA7HXt09xPTYrDsEl7vUEV8U/B9GRge
FXTNLSb9jp4LrACtNL7J0Rxs6q/tO6pAxoW76cxj4ombG2th78TG/V7R2ILa
y9Z1wKe1PqEijMwjcqy6HlbvZnHY6U3qPVzHuvK9GM74Xs8V29icYeV7f/Lv
P81gO2yp0fscjfDGbLGu8SY2zTvwbiN0vVjYO29G6l3rn2iERvm5aHFj7Nl9
55WaoCzzG6PJf2Qe5vGV8y1j01ebGG2SX2Iz2ASP3D8Nd54h9RzHuZshePLD
6dUnyX5FJVWa4e7595VHEJlPiGLZDLfhrcjNI9jdpzjvrTxP9g5FpEhh2zDH
RjXDhQcDy6/2kv5afdHNcPpfn+U6YVIvqTW0GRQsXgwo8ZH1s7dcm0Hq5bOT
TtzYWq2DBs0gTOkoyV+PHb5z48Fm2JJLF3jPQurd37lyv2Hd0hzMs4rUY98B
TTC/yuNH/M8ZXE+U704TDL1yfMb8Edt46hB/EzTl22Tefo3t2eNb1whZPje8
hp5iUz7sO9cIof+ZXDoF2AypGm9oYLf/0uGSIrJ+9cZLK98La3Q28KeS+DdK
cgPsKj7Z/N2f1CuoP1YPbP7o8WUX7NE1k59rV553RxzbrUj+gxOBNVC3du++
xHPYNL87IZXw2GN6w6wyiV/+xlkBd38V/jwui02NartXBpRxGZjYQepzGZ0s
hp2X/qQc2UTm7ZIXKoTVPfX+gWtIPS1Tai60NCifk5j+Sq5vHt00SJdZK+Px
FtuTZy6ECn75T3lf9JD48QOmj+C6SNg/IToxS8i9CFCN1x63qyYW3bXGB/Zu
2vKEnkfM8HdEF9gCX+fzUIl51Uqc4fO/hEiLKOK4p8zB0Olg7FjnR/p1Czk+
hPwvwgbrXYjXXm5NgFDTjxTjm9jGpiNzyaClbc3GpIMt6PbnTRZIP5Ge0T2F
PWrWJJ4HmyhzvZly2FR7nWuF0Cfh9vjMTuzZkH8tJVCepuj9mIvkZwnJlUH0
dsbrMyzY3Tb6yeWgzxokFfX1C67feEW5Co56anBPvMOeHbHzqAbeec4/Mi+x
OXu25dbA0HhM82AtyVfo7qmD+kuXssQLsYveqXXVQ2LPjpC7KWR98a5qAA+1
sdvdD7G1rotENYAxLV1PKBDb6f0aSRocP3LjmJ0b9qRmbyANhArEBem3sMNN
x4ZowLhrlonnKnZVOP+eRhiPL526oYfNQ79t3gitmxy7atWwbZy6qI2QEShX
uk6B5Cfs6m4Ef4blGCMpbNUsy1+NYO7Y6FYsTPYTHrO5CdS++l5h3Er6t6Xs
aQKxa6oqumzYo8cDDzQB+xD7vsylaXw+3VQ71ATT2s83/J7FXjg8Lt4Ez55E
/jw9jp3FeJmvCQooeq8TXmHLVlevboKwym3wtQObWrp6tBFsJIdTKICt/0Sp
rBHOpVP9I4uxY9X8PRrhAJ+p1XgadsDft8cbYXPUnnMyscT8Rn9p8JP18+GA
+9iCptvyaPDSs4B38C521e9dOjSomL/9b58t6fdH/mMDxFgfHne/hs3JhLgb
wGliof25PrYWX/ABgP8M6vJ3qmMzHL6hWA9yvR6RdxSxaUJ88nXwl8ZssHU3
2b/sEeYaME4K48vehj164NZgFbS48wzLrcP2fHSJWgnBcmKGhj8/4/vv6brV
5cBXpmGc3oTNoM30XwF4RvYLHqnApjKpHMiFcRujd+3Z2LSrPnOZkCdx22Q6
nLjlvUIycHL8EbrrS3xXPCge7D55v9/gROql0u0ewEA7e1qyJbanRqprEMhn
PjA9aETi3paJjkC9x7+rVZvYt/qTHjCZZkzoqZB8xV9JnmCutD9j8ihx3sU9
odApWGXmIkG8L7nyIUj9o+zh2IlN4W47mgAPhp98fMxF5r/HqZ4MBnFvzBv/
fsLzSU3aZALNyXTv+Rlib5ppDohc+DI1PoY9unX9iXyY3vzPYm0HdjjT2OVi
0Pruvy+unsQ7NHNKoKybc3pfMbagseWnUnAPEbqpGYut1XxWuQLGLHMl3t0n
ptlergSV04e+2npgM0wumldBjmh9AdMdsp47w6wa1rOo3Io2wza+uU+vBmwn
uvaLXiT9jc8eqYX+5guz1RrYRRl9rHWQ6Hnj9vAhbIrFEe+V563Rd+lbotg2
JWV7AMwUXL8z8JF53/bXA3RsZyqN3IAtteik3ACSf4LviDBic2q51DZA5AD3
oYpfU/j6kDJaef/6VZH489Qn7EmnLQorz4uHe8pfD2NXLQ8Y0wDuFNlbvsAO
H0p0W7kfaB+VWWrB1lc1WXm++0k1/Qqtwm6PFEigwaf1ZyoF80j+7pdJNDj7
pdexJAk71t8nngYlTw1klaOwtbQlwmiwJWdi4aUf6Sf3yoUGLgHW1dddsIvY
AwxpMGI27/z7JvbaDFU5Giid9JS7fwWbk0FkPQ0yhFn/8uli01TPljUA2+rI
2gJV7FGTJvkGsB7ldaPIk3pF9HKAHkiV79mPLVXtJQAg81h86aowNmXXDpd6
+PcfuuvPRurdrlldB7/EvevYlibJ+Z+9oxam/7X+CZnBNubt3FsDg+kajg/6
sLVOzq2vghdOEeVb24iZi6YqoP3Myx9x1die8WvKy6Him4FNciI259k9AqWQ
30ItEIkgcf8+KIa0mPHpTB/sovYRrSKIULC6UWCOPUrnSsgBq0BXk9oDJN97
aCYOrho0UNEubIYE519RcHE/49vGrcS3E4sDQWv1Kf6TbP+L7zFzgFMvgy61
L37Epk42/wcoq+vRmRlsT73yKE845LppoOsd8ZJuYQgI7YzTeUkn+b46K9f7
tp/DkfrV2DQdjZdU4Gzb+eJNLvGN+e5UWLbK1hgLxx6tbPXLhjnFL/ev+RCX
zaM8mN4k3TFpj02V7hwpgNdVVSozF7EpDfOTxdB9f9HXVoPEL7/YWAp0Q0rz
nCJx6qh4GdRL+65yPoBtLNR5tBzKmNoVF0WwbZ4oylZA7iv2ux5bsWfHpvdU
QkqOZt1qNpK/KWhtFYRrvZJlnfmAj+90YHw1+Atvdwx+h83pY6dRs/I+Y1i+
oQ9bapXRTA3YPUn5EUnHHk3i9aoFy4QP0luqsSnv3FfOD5NbYjaPcrE9Q03v
1IG+knUBXyKpX1bYWwea3CXTSeEfyP3xpHA9qEzOiQn7YNt83GhSDwq1R29k
2GMLqq8Jr4dDoe6Ze81Jv5fc+fUgdqVxIu8iWU9DK+/bOw+tEZHSIHb1LKsH
HhY1k1JFUi97NLEe1g8GU2UOYM/yWTjVw5r87pFqEWzjPvHj9bDowcWvsBWb
uqDwuw5+aOtforFihw9WUuvg066ERycWJ7Bnqo/UwbuFt6/oX7GLnptDLQw8
Fd5y+h029cDXQ7XwPPG6zrNebMoXp/gaoN/OjdSiY9O+yn+vhjrlme7eKmwb
dOdoNZRuPbjhQi7JX/jvdhVk3D+u35iIbZzHmV0Jsf80k/dFYs92fuytALeP
VgcZnLE9C34yloP1JVc3CysyT3YJaxkYPw9s7TMi/bySGUtBuTLjQrYKmX+r
+ssikNlXTt0s97/5DC4VgmhS85S7BOk3xj+fBxz+o67aXNgMLN7fM6FPj5f6
99047vfxC38c0DtEp669HCfPL55LD6AKHTnQ3U7ikjO/giGn5KSrXB1x6OdV
XpCwW6clrZDkrza1ugqhcSbrNqQSC1r5WYDn+tt6ztEkf+9xOy+w9fZIeh9I
PCfAHQymv0ImNdyxafIWlCjQs0iQrrIh64M5LsaC6kiOi5ApNqVuTPcxiNPb
OOZPYxv7bXRMhZ8PV0lTecj6pTn5PPjAyunCxkFc+RYVwID7jma7f+/x+bZb
fUsR1F07pqv2AduzJW+8GApeqyWWvsYOny6pLQGqhv5H/mfYo7vS/EvB97C9
8/dSbGqO7VIZOGT7NBlkYtsof80oB3P+SPa2OBIPiFh5Pl6MoOpIh5L1dof6
K0B9TeHjeC9SP7jkUiUg5/oPa+yxjfs/9VWC1Jen+23MsWcHG45XgdCVQafB
S9gMKVtTqoD75WSjsiY2Z/GbuSpYqzbPVqiErc/6W74a/tSt0dkmQ+rvMnas
hmkprsc+e7FFF5jSq2EkTejDFz7iqZHWaujmkd6vz4mt1Tf5uhqaghWdmhhJ
P5ad76qhjOFso/j8GK6PAt9UQ4adAVvMJ2xK+u4n1RA7aXF+1Qg2Q8+qnGoI
MnBOsHyBPduy+241uHX7T7xsIZ5IO1kN1srRkpQqbKfVYQzVYFyV5piTi71w
5m9BFWiLl9K4krBVdRnPVYEytZHVIxJ77caWD5Ugw9WtPXUPO/z5ZZtKqHV7
tO2kEzZVw2GsAigTJqNUC+wA9VSVCmjREM9cNMAWPb39cTmoVczd1NfE1jLc
9qEMunY0HCo7jk07MryzDLT9A/5uOIQttS1XqxQM9LcHtvFg23QHBxTDO9q4
pjA72e9PxcgiMNtbsMVj6R22haNgIdj8oaTJjGFzTv3pywGfBLOG9Fzs0UYb
lWRYu0bKb1UicbOb+WMIvvlb/XI4NjUjTCgWNvY3ba72Jv7KVR8J0Sh4kMse
m7L2gF0QbM/UTba5ju2Z+NH6LlA3CJh3/kf6G9gqGcIup0lJUXVi/fQFC8gZ
LZ7zQSQ/LCnBE/aruda9lSL+76pqEJQVK/scEybzfP46EgH13q84f7CQ/goi
i3HQLbK0z+TFKD4eqfl96aAbQv9e34zN8KyhLhMG58Kqt1Vgj5qdisiG93Sh
Uy/iSNzHYCkPzPdPr5MIwTZ+9yKqAKZjyvsCPLCpEipsRfDLXNWEYkriindv
FIPri417E/SwZ+98ZSkBBrk3M/Oq2EUzuXEl4JuSVnH+GOlXd39nKbCyW7sX
SmA7MV1PKIXQO0eU2QVJPtt29jLYPMTAfn0T9lrxwJtlEKvc8aKJicw3GN1c
Bkcv7pFW+Pp2xTyxEuvKQWlfDfzswXZakDtdDuqL6up5ldgMj4pcykHv2dvX
VxOwBwR8E8vBONH2+nYvbE/ZwpX3G4tba+Z6rmHbhO+nlcMdSqx30GlsSh8z
lIPbxn2cSvuxF3LlCsvBb6z+8e/N2MZVfVHlEF6qta94YWTFsWFjluXwyPd9
lfkwdnu1/ZFySNF1UBFswlYtSP1VBnm7WfteZWDzdN/LLlv53oq/EnYfe9Ti
qFYZNLRLzqjYYFcVvl15f29/1Oi2rEPqFT92LoUeCx22iqPYNlcTFktgjM1W
88sO7NmjKQ4l8D0n7IEIEzbnRMFEMWz83LEjqmsYz+eVkVoEgvcnr3aUYjOY
L08UgtQ+5uxVj7Bnb9rZFICW5fGDNlexKW9LTueCMbuRU6YqtjFbWHw22OS6
1Y9IkPWnLwxmQtjnylMaC0P4+l4lKJIGXZaSBmIh2MbHPH/FwAi7evIVW+zR
5R/xUfA198aH2AvYtOwprVBYPuO/r1ueWEHj2D1YN51mwyKE7fns009H4Atu
Kkcs2AzyS6E6IC4++sd++g3p/4XrOsh3LinmvyDO9truCupW2++NV2DT1ktL
3gMDjqMd2xOwPef7U4PBKk9vw3kvbMos+68IuD8d8ajxDLZg0vG5WOiwWluQ
ODqI5xceeJUCgxy7f7xsxWZ4eOV0GnzKOyG7Lhfb851Kfjqwfrnb5GqPTaM7
K2UBT0gCS+lF7Nn4VXbZICpRo/5JkeRX/fcgB1RvzvXrs2EXhR+i5oH+us18
4TOv8fXVI3w/H8zzpa609WHTDOquFoCTxtmM5Wpsqd/NewvB/4vl58NJ2JyP
mYcLITokUOqmL7aWdC4qggyJTPu0G9gBjSfvFUHFs5aaN2exVUWVmoqAfnPs
36ZDr8n1EzlXBP3rGE6e3oY9qiYuUAwf8vmDvJYH8PzOLIrF8Evj2POq99ij
pgK6xbDmqz7XbDt2rLLtlWLgCnX4b08Btr7GKtNiEJF8kGgYhe2Z021QDIe6
it8/dMKmun48UwzK1s9Fn13GNn+hKV0MOuu/3GQ6gZ2Vs2VdMVwtYCs9Joot
9e3M2yKwPSu6YLsOu8hhdVYReH89qZDz/RX+Pb5rXi+CyNCr3u9eYQf8Vecv
ghRJrzaeemwqt9iDQijpSuTQSsEumgh+UwCN1nXn/P3J+gamjQXQrnUrL9ga
W8pQXjkfug4IsUTqYhsPXjLPg8Ff/nUJwiS++qpvDowOyG1NYcPWDxl2y4YP
NV9uZ37rx/1zr13Pgu93z+8pacCWCvy+OQMWjJm9KzOwKWm5/WmwrFQ9VBeC
vcDwJjgV2FkEItsuYdOoB1a+T0TCPi0Ozb/E15NoUPNDELv9+MLYCDbn012x
kSB1XqvkYyu2YGxpeSjIHGZc9yUPm6HIaFsAKGytuP49Cttz053LXqD827xp
3gWb4n58rT2cfrOdf+kKyTe85G4AWvVdjqvViJmMmnVBL8mrh0UKmybAct4G
DLwOSazbSuoVDDG4gsnVj/6blvtwfkpArTfc2qOhwN+JLXiSbSAYAgp5sg6Y
Yxvf/ECLgdCIp6tlNbG1LmVPPoIHd+5eVpAh63NYluIh+cj4JlUmbCkfs6kk
yNoWc1Pjcy++vzVJjVKh4K9au3YP9sKVX7RkqGkodL9MxV77yO1KKtCSTQZM
/LGp/q8E04Duw33Q3BpbVK6wLw06r7WH3NQl9V68dkuH3lOuk7by2J5/pbZl
wOBeyRNOwtjmMxHZGTDK/u6xOxupZz4skQkfvkQteH/rwccjZT49E6afq5wP
GMDu1mnfuPJ7F//OD2nApuTL3c6Chai8tVEZ2KIaCi1ZsGxvdDU2BHv2RBNb
NqzR3wSP7ch6tgblbOCQa+VJvYRt4ylumw2b+JzuZClhU8eZo7KBZ1msK38v
tlOvRmY27BgdFi3lxNY6zlSYDSJN4T5V8y/w+eEglZMNYmknRupHsAfWDsVm
g5TfL9nmVuzwN5tds0HGPDuqPQ+bsmVSKxvkTxt8fRZF8oWseLNBSXyDWq8L
dlZWxassUF3flDpwBbvd/11AFmjO2i0Pq2LP9rFKZoFL7WMdugS2rAUXZEKG
Hz2nYBO2VuQ5Sib0nJthiJnvxvnsP8oyYJmP54LHEHb4Z0m+DBCbpORfb8SW
eqDokA56pTcYtTKwRYuvNqdBvlptoaANdvvhX0dSYYBrfA2rLvaCwLxRCjCN
chh8O4pt4yDllgwGDoZrmxixaf5oVxJwqzA20SWf43kqJXyiwWpIxzpmoQvP
87RL7AG02GbwmjcRZ/6jRYBtktodVj1sQS4b50DoOBwvMCiAPeqqoeoLOzun
n+ZMPcPn7/izeXdwMkFOrqXEJg/E70D3QpiIuvszcr/XMjMG0bB33XynSFxF
7ywFPHYddP/CSeL3zW5chf5a370w2EmuL//Dt0FSu/9laBo27VlKoBP4Te7x
NrLGZuh1K/CA4bvOklKyxOEvtXwgOJfP/8XTp/h8MZz7GQjjx60PpjzEZrC6
xB8CxwYa3toaYQvynOcKh09MV2W5fnTg84VD8HskKMWXjY/XYVMirNUewCNp
5ohyP2wpBffwh6BqmD2lx4ttrG03HQPUn38e7hl/gp8n8jUMj2A+SF1pIR9b
/6ArYxyc3Zn4td0RmzPG4WccZFTOxD06jk3Tse2NhyWN46cs2LG1XstRE0Bn
PPKH3Mt2cv9JvfgY8lzGk9iTsNt5fFcnAuNGGfUhc2ytkvqYRLiY6b+Qd4DE
yw7zJkGJwus098U2/PspDgQmAWuf2LmzdGxKcsyHJDC2cFvaEY5tKizDQ4Uq
hq7smf+w7fRzDlJhQ4yAHk0Y+0FJzkkqmEncXh3xhb5iJhv5s1SA5qaCK5XY
483XNaiw5SLXpQNe2KLPj5ygws3ZayyMZ7AnH9Xsp0KLX2VpLxe2KceqzVTg
42c1Thtpxc+HboXBJLArvchhn4Vt7KPrlgRP1fKqTtpid8+6ciSB0OiS6RZ5
bE6TqaBEcHHQ3PhxDbZUwpv5x7BrQ97o0aYiXO96f1cg6GjX9terpiuu/B4z
5x44omwJ9S3U1qYVj0rwccciQYG0/Bfz2PolmY0xKJZzUZlRrHnFxp/yrsSg
Dat1hg4ZYGd9PTQXjfx+5N4xCyVxO3W3aLQ8zsgeS8Mu6mecf4js+y+lPPmO
LSV469pDZFrN/kL8QsuKOWVb+B+gNzlXzQ0DsQeGJq9HIe2E2n9htdhaokcy
ItFxD0vJHwKtK47Nnf0XjqpsmltFtLHbc8u3hyMpk+2X9XyxGb4U6oeiHSef
3q+exLZp23MxCEXLCAt/5qWvmPL67LEAtE7UtYZPA3sW8az1Q4ts+6Y8irEF
M/rMvNCdRW/P4vckf/3iD3f06cvg1vfcbSumUW7+ckKvu4NUVFywKaNeYjeR
VtPYsGMeif89eMMEtZfK2WePYDOozhdqIMX0SI43nO3Ys5nhp1BF9KdUjhPE
nytUryPJAKVjyB6bpm1/xA6lO8f13MrEpkw6ZLmhBwanV/WyP1lxt/atYD/E
cTYllglhhzNcCwxCPop/9svYYFPe994KRbZCOYaP+rCL9q35E4mmNq/+1cHc
gY/n4aK2B8h4zcWQRVnsopsZ/tFIc5K1zugxtqCd7NAjRH995XzEc+x2atOt
eISeVn9qWvUUz9tp9yMBiRfc2LbbDLv7r+mTJJSW1Fh0IRabqdcghIq2R2xT
DezANu2XlkhGbHZPHKYlO1cs+in0XAryvrZz/Y4r2J5rKl+moN96zumaUST+
VU0rFdmovpD3asUWV93RmIomj+7tK5nHHr+2TjQNGe3zshzf+2zF8uo/fNNQ
P9/r1VsMsGmVBa/S0Nn10nGnQrEXBHcKpqPWfwHSzjRsvkZJw3Qk/220Pec7
dl00RKSjsjFZ4yGRLkX8PKqpSUf7+sLn113ATutZ/zodpbROhioGYovI5U6n
I95Kyu7btSQ+5jWXjiKyYutTvmAHfPL/kY7Wxs3q9Ak8x/3OlX9IR573VafX
aGPr/2V6no7ycsqCbwdgZ93nyE5Hr58ISgwDdkK9ukM6Yp4KfqY6h02hdcik
o22+izKD3N34+I+EfUpDRVEHWDilsQN8siPSkEqq+auT6sQqe8XT0FBJYqbr
deyWaZGaVGTb1OdY7I2dZ1txLBWt7WFT/fiYeNNiUQpKfEfh4a/GTjjDw5uC
Dn1zmNTuw57+KO2YjDoY8qsCZrCdmOzaqciY830AsL3Av8/W9eup6JfAtv9+
7sIOn74enoiEFe/9NjLAps5zysSjmrO1Tx46Yle9vmH+CGkZfnv0NBJbNu7F
vRjk5n5Z7sgTkl++xzcKbQqJYrs5jj359D+zCJSV8GQw5R+pn65sFYr6ag+7
rpfpUSR/71C6hyyeWp5RPodtM9zj5okY3iRvd7HC9mS4J+WIoj+9+lzoj015
qXrJAkn8WVc3kYLN4Lkt/RRqYVUO3g4krm7scxVd3OZicO41ia/1O+aIZkWL
xP1/YlO5rBW8kZ/sh8W6Db04/9jx+gDEp8r37LsYNnWukjcMlVzQfiyqgm1c
UpQdid46gMIDN2ybIfuMWOTg93NdRww2w+pM33jEES028q8EW5bxx5lEJFse
fddyCnuxdaI9GXW1dJ5NZupb8c+qQyypyLRvtcArAey+e0KyaSj8h3WDkh62
r5uITQbazZge5nQbm8fXwiET1W16Y1QQjJ0nWXozC01Jn2LgbSb5TVMSOcjz
uHu35gi2cfjBXzloy7lS6r3f2AuvjhTkojzjKZtarpcrLntfrJuHlGwEjn/b
j001NpnOQwMeuhv3nMFeMN1gm4+sw+6/MzDDFnH0/ZCPmJIaiyO9sOUfRKoV
oLiCea/2BOx2Gc74AiQFEtrLldhFl18PFiD6s6tCh3qxAyKfry1EBsOPvt/4
ip03+1qkEP2Yft6UxNqPz+99E5KFKGhxTdRLEWyn6Pe7C5Egh/xVdgq2+N5n
HIWoYrvtweOXsNMCkkYLkPq+LEZHB+xRDsOUAjS9oAojYdiH9v28UICC6VNO
KtnYs8/fLOUj8QdBBwuasKdVdj3IR0+v7PvKPYRtrDzBm48s9ndmuc9hF00d
isxDbEtWVyfWv8LPuw7pP7kop2PdDg1R7O6ZJe1cdDq2YKDsOHbd24aEHPTp
mmYU3yWSn5TyKhsFHpzV8LXD1ql6yZiN2rukm89nYg/1rpLOROYJPe61NOzY
PJsDGWitxR1Z4UFszofbd6ejU2vK879zDOD6AycmUtCHHl3zi7uxO6cvFScj
P+ovoSZF4vsTNlREPyYbG2GL3e1jFROPzFhfa/++j63P7fo8Fq155bzuSjp2
UfTNbw+Rsm2tt9QAdrj48+kwNK5ooBD7DVtwn2NFIPJdtzT/j+01zj+w7ZIP
EnnzuOS6CDZt4Gq1M2rOQjefK2B7Rp+cM0cmDm/3HLmAzWBV/FIcrVb2HEu0
ITY99u4GStm48zFzELGTa50TUnrbeME6lbhaQNgbjeWZbOqvI/WHmFr9kbcL
4zOFfmzBtRuVQlAj90mldayDeN521fdRyPj9xKKdELZNg/61aMRQ7Fc5dAxb
ajjzRSyiqLeL51ljc07zmz1Go9tufNwcgF1k8Tk4CXl+ZE1xTcaeTUx5RkUN
3me2nunDdiqWqEhBhlrTL0q+YE9uXRuTipb5Q4J5Wd7gehLJFmkIVXet+nQU
e+ENZSodjfjdqjt3Hpum8CUqA93V4XSstsLm+ZgqlYl2CBVL7/TDnn2j3JCJ
6mfOTQckYcealytmIYP67xmzVdiyu0eLs9BiUNQV/R5s839l3NkoQf8QH+0z
dtWhvZbZSH73y/49a4bw+itHSrLR0A/7iLAd2Pqebz9lI9fGLerzR7ApM7u5
cxBvWCWz0TnsSTse6RxUY6DfSLfAblduQTnoothvV0lfbBojZcV/5h/JRD8m
8c6M/TkorlXu21IFts3hbZtz0NGoN7nXukk/7frJbNQT8a1AYwp7we93cTay
CmcpObx6GO9/lZp1NmIO4y/n3469wP+WPxtRQw5WrTmELWv0k5aF5ILVar+o
Y6veol/IQr1BRvDyGvasSOxYJrIKtG+sv4s96l98JRMxB9xvSY/Bju0zepmB
qH7JbSFF2FTPOYUMdPReZYf9E2xj1BGfjqy833ef/IudtTpn5X7L7PW7V4Jr
ZMWiLXedUlGSx4ZX3BLYxty7C1NQj9ux4QlDbClKzepkZOV6bvSZI8k3jhek
ojUu19+Xh2Nznr5lvXK/d4yautdE1nsGJcehpNuf53btfrtirYqaoEgka7Pq
9zpFbIrRYdtw1GO9dXHuAjYtJTE7GDFbnVhND8RmEGcX8kVUi//WFKQQqyhW
eKCjN26tja4l+e90xB1Rz/V77Hf7sD11lP+7iazM4tebfSH5R11/6SOfkoG1
exhHsQ3QyAHUXXyVLX4D8QaVXD3EV/yVfQMf8W2WpqvoRpHzOh9RYqpKoiWq
KGTaMH8I29OU2e42Wl0Yxml5HJvmW9RlhzQLeDe91SDxHu0RR5SQn775/EVi
AfUUFzSVJ8XdZka8aS+vO5LJq91y7A5Zr5R20gN556rwFHpgG0ucj/REz3Ne
bBMOJv23p17yQnw5BttjYom9xpe9kHn2Rz72dGzK/nNu3qg8y3aHRzGpryDX
541WZy0J/KjHFuRfy+KDNDMDdl7vwB7NkOP2QQkZm4Xf9JP+4jf+eaPJ9EQR
zffE/kp0b3Q4fe/u5hkyz94QC2/knVa258giWS+254sX6kpV3Ju79h2OGzBr
eqHtqR1iAtzYnrqyEZ7IPEVXPGonNo299YYHKk8elWCRJHFfXwd3tDrZcr+r
HHH+f2dckSb1l9SMCskXE/jshBKSvA5cPY9N6Xyr74CmEjkOvTIizu+PuYNk
EmMOn7Ei68/fyrVBzxPyZQ/ew2bI1gBTFJTpXlceSbxT5+dFpFyiQZGlEtt+
eXkKLdfxt9TkExs/sRRBVW1fTinUEp8tFT2DbHvqnza0Ew/+ltBH4sMhmkr9
pF/J+V1XEPW7xIVT30jc+e0NC3RxaWnwyTLx9k1R1ohrbZehOscY3k/pe+nb
qGtT4ljXNmJxq8+2KIDf2uzcHmwG/747dkhJFH3qPYRNCeV6Yo8WD6y31lMi
8d83phxQhcLItwFNEo/X73Vc+Z4osL90mTg8554TEjt/9/ewBbGFzT8nNH75
rLuxE/GVVRRnlGi+Y9X7e8Ref1Wckf6dr77Xosg8fSe4nNGmu7B2kkr6xTFn
OaHOgNBgiwJsT4vK1U7IL8qQ80styb+xcZcjoiRKPrj1hOQ7bOV0QH+ylrd+
7yf19bha7VBZaVe83TiJj8WdvIOsIVFg/htZL+MXdRuJPrFOdf5H4rQfV26h
sV60Z5HjPfbGFoolShhZn3uXF9vzzMd2M8T5s6DE5zCJO2i9+Q91LN+VYT5B
nNJrdBb5smrWBGgRKz86IoUQlwBiNySuqxVURAs7ZhpDLImnvDzOo5K9DSc5
nYl/XKk2RFaHwp5E+pF+HwtLTNGo2v4XMcnESuOW1ihO558ObyE2ZeSJ8W2k
Y/R8IKGO2CH14h203iLJQKCD+OmAoD1qt7s1mvyKeM2GYgfk7aFoKjKBTauR
YXVC8kEbJjO+E+8I3uOMfj14a7mXYRznS8ewuaCipMKZ3HXYnh/2FbsgixyP
O5LbsRn0LHhdkUi55nyRKDZtEk66opEGAdeDMiTfe+yAK4rtmFkuP0FcOPjO
BWm/bPCWPUfW9/zWdkEco2HMtYbE6m/9nBH9k1GQghXpX+h41wl5zu1fT3Mm
8YlFOUckx8AQqeRPfPlXvT36ydbN3fqA9Jc7ymqHCripj06lkH7ukztskbmg
DX9HIam3336VDRLaR0lWryfrdyhct0RDhzl3Pe8g9hm3MUPnzhSJ900QnzQW
uIA4y8vTb30nNthwQxV17ajdwf6P2PjEP3YUEkCLyWCfwN4idlgFqX9v5VTi
IfZOrNZD7AZPA4dFJsj+fWdXnnmt3audpYnpusfNkOqjNz8LT5N8ihiDDWJh
fHfzzAVijUvStohu9eHDh6skf/TsYTvk2//ZyNuGxB1yWB3QCcq3AX53bEqJ
Vp4jWp3z61x1ILHPle3OiLZ5sUMnmlhOTtcFebivUp5NIfWK7PVdEfrIXH+/
kNhwWtgNLWlxyOypI/V9yircUF3NxsKmdmzaUSFOd+QislXU8CWpF9+23x0d
DeVL/v3uf/PZcbujhfmdvA+/krijbKMbqriyJ0rqL7GY0UE3ZP9UnKOT5QPO
l7OzckWHDh+4d50Lm+J6/KYL+pF4ZHn1ThK3fy7jjErWKjgmSnwgx/Nax8r9
xlZp9qgcyZ8M3OuA9g+duvFShfjvKj079PWkxpjN+Q/kfOg5Z4vyCrUvcRgT
WyMBGyTmc1njhDOp33/T1gxNTZvQR+6R+vacW6+gLD1zRZdIYs5/n/WQGe1m
FXcScd1ffRW0S+yOdHEuMTVugBONRznlqFcRf80LO4NSltyFJ1uIR8Q4LyLB
FwHcAiPEl8T9zNFbudDQmk/Emytf3ESJaVEsevNk/rWNV2+jy+sfeX5j/Ijn
ff+86A7ic0r8HcyJ7Ym0n9mjN+9SbUX5SZzSW+GI4s5kf27eSywpYLlyPy0v
MDWSwWYIOzHpgrYKlA3/USLx42Hibqg/oFovWpP4caSMO3rwHZ5LG5D6kQJr
7qLzBi2qz8yJaWoxd9Em+pNGc3tSL8Zs+i56sf+5HJM38Z11TB4o/FFfaVIo
qVe5deQu0mQcFD8Wj035TnW5i9bffJven0nqfWMedEfP+sd32JaR+NjUohsK
pnyKWddI6t0WH3dFp3NmOLOfEb/7Ee6CWLnmApUHSf19WUzOqN39z+rRD6Se
wE+KI/L7+M/V9Qep59B6yh6dPLdmbgvDJM7/pb79DmKqZbMu4cBmeLBcb4Ok
pHRvGPAS//r64/9Ksu94qt4wAOBINKRBCBUqCYlEEY9doSRZISsZkew9ruu6
KFkJpew9sjfXJWkLaamQVEiyEop+zvP78/t5xvu87znnOufDEcxy0mxYRNCX
hVltIYJ3zKJMBu3xy98CqmIOmZqooY/+UdWDQeZgQ+Yz6Oz+91LA5vf4zD1z
9IlzW9XhyE+OU8ZOaN9v6sZgY2N+gskP7XJqyhpi3+WrFYeP4P3XZOcATToz
YHgT/c0o4AqM3FeUZ8jC/PtHbrgBp1yETGEZepTtgCco3+uR1KcRVlbcGO0N
Tru2iy8/JZxOXV3iC8nJdnvz36ELOSP94cGGCiG9b+gu+s5AmCb/3f53Fh37
wzkIdswf25bLNEr05+n3CAaty3GcupsIKxfy8pDAa+j9xsXthAU2/zlNgkwj
4fXZYoQla1qtSND5zIVFR45wmYTuKRIsqjQyzh8bxfutbgsJhGtXL2XoE7Zk
8LEJBj1x3Xlta+yXl2IcBEEZt2d+XSGcfkBwVSAUcn2ZSAvEeUxKL/rD62sH
xjSvYT2PDtkXmBj9vswkY7zB1tQbJLzaB+/mYv3G9hkPMPnO/uF4FWFSQryO
G1Atz72ZasV5Drf2O0PFq6yelBdY/12hyh76tSaea3zEfDFFYWtYRz/y+OcY
xicMAwxBVia0/dY8etM7HmWwLnzeosYyRjjK004BonfyNP7gQNtulzSEhgTr
miRBwqQPOmet4evaknKVA2jbd9IOsCX4d/F3BTRrWukVgFmV/JtahAXSivrd
4JJDVJaSMWFlfeVCT0jsf506ehHzs8pkfaDtrODtG+6E6ZdDSX4w8cjxpmII
1idkUQKAF2piv0UTTt9jqRwExyoZouLuYD+yU3kwuIlohx8tJOyiOnmJBKl3
b5K/1GJ/ff7XJHiyZTAw5gFhyW1F20JgLlzUV+4lYctCJZkQEFry8Pg8SHiw
uFQkBHTcWq5cn8D1bFbNksDv21rHw39xvg7PJBLkmunbflr7nVgv2oODBD3d
qVbXuAmnpyp3B8HysVEzmT2ELbfcWvl7ItokbTxwkDCD9z81fzCUCjobqYzx
ya+bfIGc+0hHWoewwITILS+4x8eh9dEU+51ePegOrCz5ylLehOl9fK6OUMax
3zKFgl6/s8MGjAUrglfHY39Rc39TYDxwOO1KGvrfKFUTChWaaO+K0R8qu3bD
WS2VfrUGtG6ehh78NepYKnlIWDk9ucgSdNy7FMhDhEkv/xY6w2+Sgdn4T5wv
fZ7qBunRff6GS7h+wrEgT9C8Y5FCXzdO1GddOewD0wXDDaI8hBkOf7jnBym1
Dn0JezD+3eBTAKg/mFhYPkjYsoL1WRD86HHf5qCM/g2yJEgcnD/y8hRhF92z
gyRQmgg0VjTFOM3+QgiM/GHyybMnnL72LS0E4tZGJG32Iqy7IDodAvLcG2r9
QwmTrPmXQuDz7vjXX2KxXj55IASiDnLPnU4lLJnTeCsEZJTvbG0owvnWvRIL
gf5TgjK763F+AZtEEoSb5upHd2A/pkNVwSDpIOYx/xL7uVf8DIR3XmU3rD9h
foXjJ38gU2Qqn03gfMzfr/mCWHxDj+xfwoMLO6a9oDdNaTp97Q/i/CLPcXuA
cKOmlOduwsraR8wc4cWjTt0BKYwXe+jagM/rsy6aSoRJyRJPzoHg8NuYypOE
GYTz49XhydT50u0m6Nf6+YfB/d9QZ7gd5jNc377y/b3BfmLKA/svT76yBmcR
t/0dMYQt+/aluwC37O+TkncJC6hUfHMHulqA0+1CdOeTKS9wOMMYxVyH87DY
Nqx8z1hQi5wfYH3Ge8UAaHRa//RtDzp9xCcIbPxix1QHcf16o0Mk2BCxdV3J
D/Ras9UhUHPz9j7uP9h/q8+9ELDI2qkZsmaCmC/hlQQZ1pRn23/fSph0uDuM
DOW0fREGu9AVixVkMHl2L69FEt3JWUOGVX3SD/cBYV2frTFkKP5W9/WGNvrl
HiCDwS9FlmVjzFfNvB8Cy0z399jbEh4M4hUIgbxNJzR63AkLZJ/RJ4Hujuc2
CiGEGQID2oJgQUyPkhuN9bWcsQGQKfcma9MdrD+8l9MPtI+b3fcrIJzu+/qM
N8zqfxoarsH688UnPUDD5btgfTfWb57mcITpDzmyCv0YN/243wbSNS21W8bQ
zKu+nAOdGl5L1d/o3QZFGvBX6JXHg1U/8fltPS8NhTExkSc2od8GxOuD0V/N
1Kf8aNmKj1Yrfw9oD7tlCFs6vGZyAUtV3w/6qoQHpafVVr6nSqWn3uhgfrXM
aS9o4ptYbWpKWECHQ9AXHCLyefvtMP7oZI0/cP2yPmDlgf1cZ9YHQbvVdvVh
EmFdzqqNJHDrfGNsd51wbNP+3yQQOBp/eewW4U3vbBtDoDPvJPlyLmG6V7gp
GQI4WZMmKwiX3Rp4TQbRkNYi9xbCkurfxEPhzQ9/+txTrCfB+VCgmMi+8n2L
8zmW2IXCwYeTo3+HCSsr7z4ZCoPSRcvBUzivdeO6UIhOv8ixahl9oTKLDKks
mzjVVk8S63N4cpNhVuFyKmU94cGIONsQ0HZ/ItKxmTBJpC1u5X2gYG8lKw/6
Qrd0MCwMUBQ1d6BjdjAEgi7X0MOru7F+ruuqH+SdVNJ7Jko4vTz1tTcsk+98
2CBFmKE3+qsHGNQv2J4+jPGYhWZXKP5pOBWrSFg5eFOrE6wSrvLvUcP+DI7Z
F8HEbDMLpxbWM8qcN4PyeOdYA130A9ssTVjz+ClvkiGaXlglCub/RHLemqEl
HmedhWoZ6gHeC9j/yVyJFbA5fa43dcD934+ScQDrTGX1u1cIWz4tPOAC9W/v
dvZ7EhZYrRfnDps2/jEWCMB57WobvcBOw/izFRnrT4gU+QLNv/pyVgTun0/d
LAC2VmyZH47G9bz+Pg0Cp5ErZOGbhHVvPfEnwf0dz9nsUwh3yes5hwCvgWhS
QQbOezFSgQyu18IFv+dhPPn6FzI8ah0uEr9HWDIlxz4Uds6ryDpXEY4NUnkQ
Cl4SafTSBpz33uJiKDyz+as1Rcf55m6spcCulHOvVu4RYp7ohclQ8OuusfB4
jvHtKuWh0MXKOVb9krDL8wKtUNgLrh6/3xEu077TTIYgj87lI4Po3sJ1ZHhV
KBbp9xX7KcUdDAHxTxEcTePYj1lm5e9VKPfXu0vTOF+81v4g6DulJqK0gPM9
lSn0h5G192dYGaaI65uikOcDvx6otnSxEI4NiTf1BCZy29VbGwhvqsh86gqb
QNXQmpMww4nkeUfYsdgqKMZHWMDk/pcLIFaj8mNGkDD9bn6vIRxxa61rEsH8
96yJh+CYhAol7ACu9ybxlBboj9FP68hifmvejDlY5SrzcSsSHkyZXLIDZ2v6
1wE1wun9GWtdIGCHckW+FsbFdr5xh6t9LYGuZwjrem239oakRCVNeWPClj20
HD/I1mvhXGVB2EXl9t1AqGBXGnx6kbD98EchEtCf0IoSnAhHaOmdCoHnVPA+
74711JyTZOhTpakK+2Hc8/rOUBhZVmT/ScJ57y22h8JcQ/O72nDCytd4FCiw
ylsxhxSN88idolBgk3Szi+ZNwsntcncpsOOngsKWO1gv2EClgHhRE+v7TMI8
ql+UKSBnp/AyqwD3x7PhWSgc29WU6lSG+cU0kVDQHzh6SaYW+wu2GZHBOqVR
ZrmZ8Hw7r0kIXDE6yviwnfDkTlYpEgRwND6LeYrXc4GxKRCuvpBPNu5Bj/s9
8IPkaw0XBN/hftebkL0h57j8gbEBPC+e33PuULGqYbHiK+7vgYCIC9Bb5Dr8
f+B8YXKy9tDpXx+nPovXMyLN0ALeH5Y7v+EPWrv8pxaMzNSJvGacJq7/t4WY
QzBXemQ2dQ1h+vHdNUbA7FTXYreRsMB6mQc2sFnkyDVJLsLp+TJql2HHcK3h
Aj/hsnD7bDcQTz8s1LaLsOWvcxe9QN6s9sdVUcLKxxQHfOE4z+H6s1IYzzLe
EAj6vTUU/iMYV730PRisY2V1vwDhI0d75UPA5WQN3z0NwvbjlvpkCFwj+83r
JGHdggnlULjaXl2hdJZwMv+pf6GQTJIJWmNCmOdQRDQFchWqNbstsf/ekR8U
qJw/tPW2HeZfneIPA3pV1aC1M+HJsxf5wqDT5VCxmCd6THmUAh/Eq7xn/QlL
StJW7ofREWm1ZjJhkaizsyv3V3YlOzWS8J2FyoOhwGwl3acTS9h44wtVMmze
XpnDnYT17N17QmDHu4Oug3cJd73QaAgG8ZsVCgXZeJ7Zvb0BIHfm4Bq3IsIu
8RxpK9/3TSd1HpZj/pyiqRcYi9gl8NdhfHLqrRuM3Ajpc6URjtVvOeIMPv9S
BB6243oP9urbwhrHGlv+p3g9Xw/wmUHy665i12683qkxbRogovp9uuMNesrh
x1GoK1ktx9+P15t0ptAYTmwTCHYdRtdXvbGBtxT5Bx1jOG+M0aIT2E/qr+ef
Irzmsf05N5g3vXLG9TfuP6nzjydEPIxM6lgiPBL228IXeKSzP/IxzxD359b3
IQGQn0rb5bqOsMvf5+eD4fC6dw4dmwjT+VXmSdDhOVPKx42OD+Ujg+GnDXMu
2wnbt9hwh8LXkyIKHbsI32t6PLby+16nSuYTJSx6sSOBAiy7zz9ykSTsbHaX
MwwSY7zZO2QJs/PQHcJA+E+cPp8i4ezPnTfCoMa2+LaLGuG2seT4MDjW0zH4
QJPwsiK/bRi8VvwkzKdLeOxdGHsY2Bb8cXIxJLyjQiqKAnOcXJUPzAiLVxUO
hAKVJLnAe4Fw4YvCtaHANa6l5OJA+OAy73oy5BpdDHtwhbDhCY4vJJC5H/yU
14vwlvRqqWCQknkefV95dsWpX2dYyMDwg/nb/DHCH55SXpGgM1tBSeIU4ROD
R5eD4I6ZR9KFs4SjsuzWBcAlzuKfyecIM4R29vjAkWefj3daEObpWrfy94OV
wpe+ypawS9rGPFfInrmq6+xG2PLRr8MXwK2oLT/LB+t52eUNQPnCIsO7ILS8
0CoxYOc7eI49jDCJX6XmFHzscShXu0Z4sEdCwxKKr2as9Y3DfOi4bw9+qu+s
7iURph/1vOICXBUntmzLJmysoTfkDcMOpEs6hYRj302AP1QK1rWFlhGWVPA0
DgLyu5+89TWE9fNrbq68b8ftdZ9oImy226xv5fnStHi66z7hTWcfzZNhnCFp
17nHhHv3PvgZCg11nf7RLwjPH9ekUSDShaX3/ivCTpw+F8JWni8QX3iP8xZk
9K/cH4OeFIkhwj4jqgeoMJtU8uHCCM7vu9qACm2nvxy6NUH4SHfLaSrEsm6/
3jmLlvXcSQWLFv0vq/4QLrtv0B4GEt5RinKMv4j9OtMgDP5KtN90ZiXMwy4X
Q4EnX//8yNpAuMskoCYUbqVKH3vHgfmzjtVksDd0TGXnJRx7ied6CMiyZ82p
CRBWVmgBEjB39On4ChOm024mBUJP4Ja8e+KEdak5jn5w5UeI0TY5wpPc4rru
ADn1pTpKhAedw7qdge38FCtFgzCD+K4kW+jj3GdZr439ovfOmUL+M8u6iTMY
j6zdqQ5elORNu43Rgo86lUFDocv+nDnO83brdRPgnGVtjbbB/gKCq2xhqEhp
W/slwqSZHHNnKLvg7brgQrgsP3ivOwTxlT6W8Mb5pi988YKTL78K2gQS3hSf
Z+MHfNd2+N0KxX6FJzMCYUzVsKczkvDI9zfaJKhbvC7KHIvnx62eGQLUigdk
uUTCLnnuPWTQv7TU53wHz6+U/30oCAnJSGdnEha4LtRIgal3Ttfe5RO2n7jp
uvL3IS77M3sp4fyXVUthEK354ah6Ne4vbsycCmaMnAm+jdj/zYsEKojWa4/f
ayUsUimaRoV5l1D14YfYP2EikAoPRRrvbOskXKejIEWFk7ZbZV/3ErZMPFQS
Bl1ZV17Ev8f9v16ao4DBp8f2p4fQV703UqBvx24mtlHCJyTov8lgYRaU8ugn
3h9aMuUhMHzr7aGwOcLpPu6KJHB4c7BTZQnP70WxQiBMcF63W141R/Rrmlp5
H3fX+8bQuI5wV5zDc0+Yj1G57b2ZMMn+WK4rBD5PkT7EQ5hu5+pzCZjWzz2b
3IHxm/2iVhB+Qte2ZA9hBoHG9lPARi385yD+v4UlZSDuPvMtYWm0ueCxc8DF
aHHwsxzWZ9+Ut4UUqH+apozxrGTzKyAQwHHR7DjhwdENjB4rt+flZR4dXP9c
DJsPiP5+mPRKH+vZTz/zh9JDQlLxpoTTJbSOBsMht4AnOtYYPxMkGgL1pa8v
rHcgLBlddpUM8ENy6eEV3G+CT1Mo3Be9lkjxwv5Jeiu/ByfsvxxQCSTsAovR
YfA8R+nxUihh5cJL+6mg9/mWdcNVwrFe/clUeCMw+8crDue9/6Jr5X4w17kp
nYz10vO9VPiUki8xmUq4bFY3hwq275geFefgPPe3qFLhO9d5K4di9I/qojBw
0a9d3FOJ52E0PUSBX3GbE4bqcf3QlpGV75EXjvvT6OhofRoZ/rF1dJg+RCuI
XAwBipaAJU8nzvPjU3MwrI3wW+jtxX45mqYBEP2gNz7uPZ7X7plcH+BYdUBc
ZwjzTR/neECycuSDdaOELTNkz7vA9qDP5g9//n8+mrl2kNmoOB86h+d7a/aw
GexdSIpTXsL+fOo5ylAsOy26tOo34T+ZcidAyuNke/069GuJPxZQU5573msz
YZJxeogDHP3J8PsgD8a9rY+5Al3cNPbnDsL0GaFIT9C4VL2veA9h5WKpUF94
krfxvr045ltKHAiE018czPZIY79j/Ook6BVq//VJjrBL0+/2EDhnuSMmVZlw
GaccTyj03/URMT1OWLL1wMr3wIX3Pa3cOhjncTwQBiM8+0179QkLXGyYDIPL
huGzsabYb0A8mArTNz5dP2WN69Vt7lv5GOk+unedA+FNW+cZwuEveyK94wrW
85TOUIF0cvJcqBfOO6hRQYXVV7VmlAIJx8asV6QCm7cSa2Qorn/POSMMOG0O
8fVEYn7X23cU4Duz7wBfLPZnEPy28nsFO9RsEjH/hc4TMoiKcRiV3MH9WN6m
hIAUzxrHuUw8PxMnbhIcWb0UpFSA5/ejcU0gKE9PxUeU4vwLrrdX3scHvuZ2
VxMejPr5zRN0nr1v4G3C9alXf7mCYX1X54U27Nc1E+MI53MfDBU/wn5Pih5Z
g82NhrlfnZjvZ6F8FpxIpeuUXqEFnHmnFN0vZ++IeI+Oc3mmB34mtw52f8J+
a2NvWQP5ePQx3hHs5z/i6AiRh0JNLkxgfvf2UVeIFfRxLp4lbLlm1ydPSGa/
TP61iLZVu+ULaX+sEoFxnsjv2LnyPZE7YlgYzkp4k9B5fhLce6VN69qA8X1H
o0Kguk25ZxsnYYFzUb1kaCqV+WrNS5he/3kmFNrviC4WCaBXrV55/38auZP9
lzDWJw9XhEGPF6cQ7CdMoubpUqHvwlrZcGl0TFbryvOuu6zZJUfYUtBidTiM
KM6c36ZMePB8HF84TIqOuFofw35bhZjDYZ77Y1jRSZyne18LFf4x99ya1cN+
T6ZOUoFluqNE8Rzmz8ys/H1gH2hspVpgv8m+ld+Hrc/KXr24iOsJkcdDgb8+
Z5THibDy3V9PybA79/aSlRv2l3Yjh4DYjZjNRT7Y7/LFzSQ4SKLsmQ3C/F3v
ZwNA/rKvnGIYxhUrg3zhxPELVi/iMH6O94kr6B4y9uRJxvn8bu1zBCPBU5FW
qZh/je+sNZizq94tzEaTrx3WA9s/suUzhWjLjD1scHlE7IFCObqVMnUWPF4J
vAurxX757aIXIKBt64/OZoxLfVh0hNDSdYw87Wi31Qfc4Oqdf5xWTzC/e90e
L4iPnBUp7MLrJdvW5wu3vEYVZl7jvMLd2oGQfqFfV+Ej1hvslCdBvu5Lm7DP
WF8slx8CpYqPfDpHMT/vzgQZakSbo7gnCafz9K+nAI27It1yDp3SvkyBB8x5
VQV/Mf/68Y4weDaV8miaaYGIb4qxokJvf+yHo2sJW3LOvqDC+6dhk5SNGFdn
2BwOyXBXw4CLsDL3n13hYFhedXvPdsL0FlXGcODc/eznr10Yb2EqoEJP4mf1
DlGs/1PAt/J+ufbPrUQpwqSysXNhoBOw5aftEYz7NthQgO3nPvXDStjvpJ5i
KDyxUrnFegzzgyW+hEBEr/HEm5PoJnZzEhw77qKWfxa9c/xsIDA3hCf7mGB9
q0GiL7SJp/04YYXzfLwc4QmktBrVbfaEGYo6D7gCbOlMGnVGix7c4wB/KV/G
6z3RTC8/mkPD778qVwPQucHiGuBziTPJJPR/F5HVQfaj2LjoVbTXLNUcZk+r
qfyJRYevX7CHijaTxKdJ6KCQba7gIuP2PSUVvW6tmydI5EcqO+XgfO/+BPvC
OG/GTYVijM+3KAZC4fW6MbZKNKfgWRLYM3QpfazH6zPkt/K9IOz+LaGEjr40
tvL9PvxleTTwIZ7n31ITCmQacynpdGJc4bRuGFg+3Z+w4xVhAa6YLVTYARqj
E+9x/Se+6VT4UGYGLUOYP2C2QIXbuzxuxIyiRR/zh4Nx4rURi0ns3x6/Lhy4
1mYpSv5Gc3W3rdw//g3xDMuY/6RfgwrxE93fupgXief92bmEMNC1GlXIWE9Y
wGBXKQXYexniXbcQpr+ySAqFZ8d4vqlsI8zA4q1Dhqv1BxS2CGD8LcdrEpwQ
Px43JIxxJk3LIGBJM/9asR/jPVXNfkCmXI89e5SwclJShBso/87+sksV81+P
MDnBskOT/OwJwqSpTg9raPrwMqb9NMZ7Okd1we/09+EEQ7QA600JONLGJH/x
PDrfTtUY5g7xxsjYYH3hk8mLUJUnNbzaEeP8nppXwI1XU+61K87D0MXuAZLX
LaNzfXCe89aSPjDxz/uzVzDW365aFQDFbjFHjlPRP2JCg+HSl9zr3NcJp3/O
MAgBEWPa0LcbhC1H9BrJ8PXJq8N1t7H/5JqpUMhW/BEVkYH9GWR+UcC6jHnI
OB/zxa6tPN8Cu/gP7yvFfreZbKjQf1M6aqEaHfLqGRXurNH+9LgJ5xcKWfn7
f87fWvb2ffT9B8sr1/8uJ8vvx3j9nk6v5Pe1dLw624XrGz12ogLvkE9O2Wuc
/0nlyvemCbOY54aPGHc9vSsMbgt/VL/0GeuvOR2lwPsTMZwPRzEf2EVWfu8d
VYZ3TeK8Ny+NhYDZ9ZlK0hzmU8ZCSXCnNCf041/Mbzs7Hggfuo3Oyq/6Q5g+
EeYH57mapmc2ElYe+T7iBqlHnNt0uf7g87Xe+zL0mwjEl/Cj4/WLbGBHYI/V
ul3oC10zhmCeRpGy24ee1B2XgNRWWcb2A2iepkYdGPg80iUgi+tlHpu3BAGW
lPRABYzfaBp0AEuRUy59qri+uauEK6Rr/VM6rIn5q9dLecKgU/nGhNOE6aA0
4wMCMRcGJg0wPzLcJQCsyreWnjLDfu/bsoMh8+XDoEJrwgIveI+FwNAvXx1W
B6x/2x5HBiEe8R02V7C/6LfKULCW7/9B98T6bQpFK8+7WWzz9gB0fplXGHwO
Ur3uR0aH2mylwu6MWbM3EdhvVSGVCjb3c8UPxaCFWZ5QIfuL8d/Ym5gfJTlA
hS+s65/9SMH13Dk6qCAs2pyilYnxxqdBK98DJ6845uWj3+SvpUKus+BR5lLc
3ytxuzD4GvtynVU11huwJVNAuDKsr7kRHS5xNxTsXh0u5G37f/0SbzLk/x71
9X6E9UG9IiEwsu2OZm8n5hccswkGEQWdbVKv0As3RQPA3pxh9Pp7XP+qeb4P
5JMq6sY+/X89n33ygNFMm4jjI5hf4fbeBUQfcBlnT2A8MuqQA1z69mgv4y/0
qMZxSyhc6//7/J//64+PacN3sf0PGxj/EqYvvjoE4joDidxr0G0l80bg5BJn
68GOlhBnvgjF8Wqy3ZxoX3LuZRiv+rVagg8tLz3uBpcXzuV824tOCbDyg3t8
bJ7qEuhVb6tWvm8VaeoZh9DCjBYkkLB04VyW/79fU1MIOJOFhk1UCCuv2jhN
hnvZvZW1xwmT1t9ZCoWfHdRQTh3C9K1y/RSQHD1y1lUf44M/4le+z9Z/F+o0
wbjVLC8Vfvo2MK+3wv7HrDyo4DwS+fW43f/rT2VQYdzw3CPKZcx/5ZlFBccH
IoWt7ujAnz5UGJOev7bsi/1TooSoYJ/58PJREsanC9LD4NumpNM+VHRCzwwF
bINtpaqjsH9vCz8FvvyQ4ZiOx7jaOf5QsDFb/UviFsZrq2ZC4POT3teOaf+f
H1c2Cazlsuvyc9BbBs8Hwac899tfinD9Aw/E/MGSSy1AqALN7FvtDednPind
oaFtY9hc4INVueC7dqx/MPPADky6SKu4nqIfrd9uDu9A94teNzo/f+44GJfs
fBjzBq3bHCgPb/h+5j/7iDbmbzUGw6u0q2uH0V9t+C6Cvt15ndBJ9OiOCjd4
+Upckj6H67foXvECPfW/m5f+oncN//CFroqnM3Krloj802/2rXwPCqa88lpL
mH64bSMJOmMu1VZuxDi9a+X7+OSy3K3JrYSVxdNvkuGp01r//fxonW8rz5PW
+7dml4QIk7quBlLgsWY+5IlgXP2udBicqPMWGJbAfmssWsKgQ/g4k6AMxvVA
gAoaN7mGzx9F73tjQIX2VV8f3FbBeQxnLKig6lad9+Y41jtfBSq0DlIiOXX+
n691KgyUT+s7ntHHfNNx3zBoad51KtoE+7Fce0MBEJ+ReGqJ+QdPrryfNt1u
27TGDvNXsXGFwtG18dPqlzGu2TsfAg3eVr0h7rgf5w9VK9+DXyVraL4YZ+FI
CoJafYbkP8Foow8B/iB7/4XvESrma97c4QPVUmmmnlFon20kDziU7qxYEY+O
SstxgQp22PkzGeu93PY6gFTgBkbxNHRMbqollH3/MGSfg24q/KEDEibF7TlF
aEsxtr1Q8sg/d6gc7db8UR/ED2tH7KxDU+x4L0BRDu8lMxr6a7C9I4hyjmnf
asf1d5HyXKGAXL//9ROMSx619gSRqYiNHN0Yr94y6AN5FsZTp9+gx5i2B4Bw
596XUR/RGdZbgyFb4XfV48/o1KkhEuwq6khkGUPrPtpKBuO0uvonk3i+yc9l
QyH6RuGH6N8Yb+0TocD98Dv/9JZx/ebWMQrM+0cLca9ext+TG/5hIOFC0ni/
nrBy4d2+MLhg42aftmUZn78rrFRINra5dmEbWuglIxU6Txre2yuA9cK9j8OA
WeVE93dhrH/UYBEGcjLys6X7MZ+JvZ0CzvvEuT0OEabvvPF75X1m+w75I0cx
XnLqDxn6Nm86/1cF672Cu0JgIwsTiX4C+/vm+5BAfXEmk3IaXbHheyD4TXx5
cMIQ69v3x/tB2dCbEbbz6NE9373gy+vH67svYD7Tt0U34H3aKHHzEsZXNV2/
DLotJWfOuaLLvhXYALUyzWO7D9pS5p4hNObFJX0KQjPYSonCZEpoQ04YukON
VxeEYz0/OkShC1XOW4EZxY5B4gbaw1TtEsT5nNs1fQvXf3bynCt0OGkfq0nH
eHP0aU/4a6no4JeHzvy37ANSBgei4B76W651ANhqCpYyVaNVWL2C4Y4iR09H
I/bTieENgW6p1b+utuF5/jm48v7MIvyb+/RjjFfZ64eCAu+oPEcXxv99E6OA
K/v7829e4/m+ynlJgVym56SUjxhP6tUKgw9ztCyLYYxvGI4Ng83fyzp2fcf1
tUNywuDYQObotymMNyRGhEHAywS24nlcj+GkfBhUPKQecPmHcUaWRgp8a/TR
O8Tyj4g/t115/vnLLnnOsxFmuHVKPBT0ss2SmzjQta93kCEiWaeRxEtYOdJ0
mATNUcr96oJYPz78NwimSQcZ14pgftfHj/4g4rl793MJjNu3XfEBcweu43Ey
aPX9tR5w4/yaSwYK6PSD9S7w6Mxi1DY1rH9Gn7OHZY3x0o+aaIkAVks4JN/f
k6GL9rjYoQUOEl2/Lhqh39ocOwypQm08ouZomHM1hl6uqqMTNugbJ0Quwrr1
ueYVjmjGnILLoPQvKcTLDdc39H3rBh4zkdnyvhifI9/ygsJv/g+Xg9G2PVv8
YOD95bE2KvrafaVA4Oyy2BB+nTCdupWdBFrtZyS1EzC+R0wvBILr1M5uTEHH
h10ig6xZ16RFBp7foZDjoSvfH2bRZXlYf1/lbyjkZo2KMd7DeZxXR1DA/LjX
4zNVmL90eOX3gOs7k11mA/brdhMKg87oGOYZOsbfBUiGAfUgf6baQ6yP0+IM
A3idr5TwHPtPLT2nwJyvzMfhl5jv/PU8Be5tb/OT6cP4N1J7KNi26vBQB7F/
gTdTKGy/+L769VfMN8/kIUPvGvuze39gfA3LmhCIKp6d9J5BLxdkBYO6bkj0
owV048eJAFheYElkIz6GVu6/1llf6NN+PnxoFWGGPw/rvKDm7g3p8ywYl2wF
d4j/eY4cthbjC4Vbr4CzikB3CRt6bKOaHWjd+Lrz9Ub0nTc7zEH4S7Hz8hZ0
UzWnFjAddm8W5kKLxIzJQ3+EHNvpbWgrF1FLaOj7Z+LNj57WS3OCRPGOgrSd
aL33Bh7gFhQ1/1AI5+FgEfADnS6945N7MN7TmBsMokLbEnn2Eabv8nBdeZ49
BoaVxQkrW498pcDQgxxphwNo9quS4dDM7USOO4j99A20IyHZ4WB3vQz2Y3c/
dA08Gud3Dh3B/OYjo1Ggu6HFeZ0C9g8XsYsGcYuw5oNKmH9gqigG1pRrs5mq
Yv5yjUksDDNtMQ3VIJxu/ZIeC3T9twVFJ9BTdqxxcCc3df6lNuFBtpz9ceAz
b3P8rw72O9N9MA70tcQSd+vhfJGaXHEgeWdq+KQB5leT3sQC20SttKcx+tc/
z1gYUQoi3zXFetHnQzHQHqfe/cAc5517GR8N6Z/XCUxY/X8eRX+iIECm25nr
IjqDa/s1MA5PagZ7rOdc/TsCDr07z2bniJaLj6TCJrHdpjHO2C919atQGA8Y
K6h1xfk+9bwjwaPOsvkBD8yXlpbyh2wB7+NrfNDtWrs9gOSmmCjp///1rdnu
AKbtq74YB6Ejyp6dAVmuJ9IhIej5s4MWsMU+llxAQV9re+MDE/WG3d3huJ68
wSQFnqzfLrB4FeMZ1eTrkHv+s7NQNO6nBthvALm0oFkrDudNLzBNAvmzsqYp
SVg/2vk9Daa+XddiTSFsyemukgn5AV/k3FMx7ixLzQbu3AQe7RyMb/WfzYdO
+XHW2nw8b5kS3iKgvlD7LVSM65X6SJfA7Pz0q8UKjA+6pJRB0XWtB7Y1hGPf
nY0pB2uhzKqeesICPUW+FdCtfeZGIZ2wJLPMgSqIGMwnc7UT7pqUWagCZU8G
N/JDXF9Vu7Ya7qWW6pp0Ypwje0MtXJRmVe7oxn43p7Nrgf+R+YGDr/D+8neU
qINes5odqW9xP625BXVwbWoD+7oPWN+qxlUPqtSLS54DGJcdda+HRd7m8U9D
uJ6sRls9lJdyfjj1FfOlNjA2gL2609P6UZx3dulAA+x8d79hzw/cX+p7nQZ4
fZmvMG4S58l3Pd8A15ncby3N4Pn0xZs2gHrSkwiH3//f39s0G+CPmJDPq0Xs
5/11TwNU0H3tVJYJl63+PFUPlwy6DUsYGYn+DP+K60FwTOTYttWEGWqkjOrh
bRBJJmwN4U1NvpN1EMvxdvfUesIuEx986uB4/gHO8xvRl+0na2FZIXzV4y2E
JW9LGtVCdXf/9CEu7Md9pqQGnGxlh9K3EbZsH5+uhr6YL3QfAcw30ThVBXG7
FcuGdxGe3OtkXQkn6hPSdPfi+uOf7CqgdkgtSEQC53X+qF4Gzt4plxOkCNPt
N8jfA2G2GTMGGcLKc9eeF0GCTObRt0fRO+UyckH7yYKouhL6uuDnLGCyOMNb
poq+U7A+Axpm8tfyH8N+w1mH74BrBMNCuCb6t96+RBDZbjwyc5IwKU7mYwwM
lpe+sdDF/ellc4RC0jHWh0/P/n9+1t/tQOe9ec1hI3T020Q3WO1Sk5NlgvWO
k5OR0MTMfnOjOcb9DIVugphEs/s3G8ICzjHembAmhE8qyY1w+ohaeBn0N4bs
m/HE/FFdiQqonvsmeNoX59uv3VEJ1k5VW1hJeF59Mx01cCSPd/2FUDzfsvyV
+3XjEGlVCxX7ZxmF1UOT0alZr+t4/QTmlxohPr5yvCcW51vctbMZHJ5v+yKR
QLjsOm0/DZTXkD5eTcK4iOe+FuBS+/rq623CRzqW1tBhPPBkp2oqzvMrzJ0O
bXUVHakZOI+7dC8dkmd4Whaz8fq1j4q2whWJ4FrDfIznxni3wjGHL6UVRbhf
1o0NrcCfrZ3PXor92AymW2G6vzz9UgXhwXV6Am3waBvPrY5qnCdyQa0NUvWD
4oTqsV5J9XwbeMYMRwY1EebhFHZqg5NPtMh9LbgfvkyXNhBaXe4nex/P41bN
pTaYV+J2j+/A+KKtSRu88At0nHhM2KevXqkNcqs/X9B6juv9reNtg4BJTbPc
LsLJOq7fW0FPrEyfqRfX+zpd3gr7bLlOmb/B/gxnnFuBISNAo6EP98uQKdAK
r98PKXL1o53/PaJDMZemrNsnwl1zYXZ0CD1TKtE5jPt5WTjfAiZRW/eKjuD5
uQlubAHJh/47qd/x+v5s3EIDFqYh7qEJ7Ff2bFUzfFQ4sQmm8byMi4ca4WoF
J+PcPObzmzjWg+UPv4Uzf/F8M+a31sFhkU9TJf9wvU8sZTUwfLdk6CILE/G8
lhhVVQJEHmt/xcGEfx+6eQqBs724SYobffJnSC58X95SfZ2XsIBDW08mtMn5
loxux/hpuW2pkOwxkKMhSJjhQZ1VIjiXaqRm7EbXl626DhpjRYlLe9EPuaZ9
gX/PlphzYugI0xfGMG3hE14tgR4rliLDo9v9wZsPomvMQ2Ih7ZW6z2UZwsrr
X99PgpPamx32KKA7OL9mQo5sodZNTcLpGpdeloLC4/Ok2pOEJdNJp8qh13Rz
zbvTOG89W04FMIf4CO4wxLjJ8P5quMMhbqR8jnAXs595DUjnDkRZm2H/GJuQ
WrB+emw+15rw4KnlOyu/9+cX9z++SHjyOd/NBoifLLnw3R7Xb7sU1AiioVa3
NjgRLosvMW6C1q1bXxy4ghYo2tUMxvmPmPXc8DyfSQw0w6R8gLyHJ8Yvr7lG
g4jnB1wSfXD/Rmx7W0DA8nNOnT/Oo8tR2QJ104nv+4JwHu1AbjqcDtPavBSC
/X6eOUWHr9zLx3aGEa7LWvSjQ1BheYBKBO6v0imdDlyKFysuXCP8Nj2tmQ73
XvCMhEVjXOh6Dx2OWT/bnh+H64tIDdDh42zw2ScJhHVfRn6mg0e4dOR4Ep5f
QNJKnI33G409hbB9idlLOmQX356VTCVsvLufRgcFJR3RsxmEj3QJZdKht5vR
0jObcH63dCAdHG2qbybl4X58+HXpwPTb/ml9Ic6vMcZLh9uR/IwfSjD+/Fxj
Cxzk75JdLsN5XMclW+DxvVAngSo8/4iqRBpYqhzOVK3F+YL6x5th/uXYG5sG
wi4cKdLNEGubuiG8GW2x07EJRBbOqBXQcb9Ps280gtGO+ns/OghvUrlXVw8T
ZU7DG5/g/VPWUF0HVDUB3oPPsf+adyvvDzX24WFeL3F90jWTauCsNDD7N4Dz
tXrvL4NijbXxgp+xnkU7sgTU3jY9VPuK63XlZxSC29Iu6Yhx3N+uvzeyYV3s
G/vCn4TpOi+uZECm0LXUZ9Po2Ra3O9B9fHrt5gWszzq+8j5t35ejJP0XPaP7
lAQMl895GvxDP3vQYATJ/9iKvJlWEb6/h/h/UTx98NZqwiSbhXgqPNrtwdW0
hjD90A23OLCo3Xuyfz266l1AEkR/iK4V2oL1brTEDGCseHk9YivGHXiys1fu
Bx6bCR7CyupsKXlgcjBzU8NOwul/i1VK4Dnrt687d2H+OqefpaD8Uaw5TJiw
ZL6HSjkIR1Q7nNlPuGxioKgSbp1fVKqVJKy7+f7TKmCTVuLafoiwy7HpD9UQ
vIYyTj5MeNMo10ANTH981DYij/nTA90r73eVG27pAOYvMq5cr7cReleqVHCe
g4cj6kHLPEmDVwPzLyppN0Cz9Ac+0gnCXZMDSw0guVZw+os25j/8nNYImf0X
H2mfxv078kk1wdaqwtRyPfR3s/ImiIj86cFtiPv/FCrQDH/MD2kHniNs6eIa
1AzOh3wFP5vheSuxPW2GT2tpv09YYn9fZRYa6A8wdd67gP2e/DtAg4dVx7M5
7QgLGKucoIH81Sg/v0vY7yHjaRqUWHTrDl7G85JWU6eBgAzX3mOu2H//BhEa
3FhnulTkgedd57TYDCyDaS83++B5aHo0NoNv9XCBtz/2bz3q2AzjV/eRPgbh
+k6Da5vBwtLZUI1MOHaZktQEPTKV4gVheF7xZzibQGP9PNPGSHS2O6kR6gYV
3nlEYb/NOz40gFhNSGlfDF7fLZF7GyD1WkeY8g2cV2Hauh4osroH2W7jeXN0
FdbC7/U317jdxf0l9tXWwKVP7/rfpOP1HWGvqQbdqAvXsvLwfFlNwyvhvlW+
1doiwpObpU0rQPbwj8NX7uF5WlcLlgP/kNewfDX2c/+6qQRiahsb0uvQ2w+e
LQSm6wxxLE2YT/IPyoORw1cVe9oID8bsSsgA0w0vOI504HnG6zPehc4hjrG7
j/F8kpeqEkGlzpi+6jnGN6RqxELV9buJDl3obycdQ2HvhSGnFy8x/+CAgj3c
PrJXTeYN2jj/wSXYwO60LaUP8+Wu55GB9LnsJ0P///fHButomKn79cD2E7rp
+EQC2EbL33k2jNfzr0vkbdCWaz+R/B3n76flZwLJ1tCsfQLj5E1bcqDqxsiV
ySnc72M1hzzgn2BLOjGP+Vej+4tAly+t0OMP5t83/lMClBNStPRlPN911Yxl
MJ5p8GWemZk4jxvXR8tBoOvb/O41hCWPvHpWAfpLvmxn1hMWGCvNroQIUTaB
QHZmfN6GXKqg2ShVumAz4U0Bh6WqYZoiefwVJ8ZV/b5Ug3BFmwkjD2HLAf/r
NWAyoO+8n48wPX3TvlqIZvsWcm4HYeWYTXW1xPvGzTBBwumORvJ1MGe7vqB8
N+Gy4vbSOhBNuNv0cS/h2MCjPPVg3nqga60Y4a7GzJXvs/iJ1s8yEoQZfg23
1kMHn/5vKylm/P2bZGqAxRNf10UfwvWma2Qb4ICXz46Gw7g/SYmV7zGbrHUH
v8rj+kanvRogueuOxhYgrHuCm9wAz5YkzoEKrhdGJTUAg1ir0yV1zJfJcG2A
Q8ZnSYnHcV7+KwYr34NhX260aeG8d/+INcCdCu+8iVN4vrUGv+qha2BtI+8Z
NGNseT0wb7jTeUwf5ymiW9bDEXmJITcj3I8f46p6cLKj/0o1wfPsuJJcB+kJ
emufnsd8P3HBle/Z1mH+35a43gazu7Ww5qeX5C4brA8XXvkedtVMMfK/hOst
Kj6qhlyv/Y55lwlP3hvnWPmey2oJeumC16dx4WwVqC5/zhHzxvWYtlVUgJeY
V72RH+azFXSVQ6HxmuehgbjfDTeGy2BLpfjs+1Bcr3P/+xI4NkhjXRP+/7wJ
HUXgt+EM36GreP6HuNIKYMjOUzUqFj3Gy5sDXDdZDetuoG+voWWCVtsth+FE
nCf8y5l0qOSnxSrcxf4HeG8lwoimbrZ9Otp7yiEO+L2HahOy0AaGqhGgm+3x
lJ6LThBh8gdKN8vAeAGaYfyuPtQtJ0/zlKCz85wvw7iYGItGGfpqWhcJBM41
b3OtRDck8VwFferp/Xdr8H5i1E6Og/xzk5r19RhXaJy9CX/E42xfNWG81/Hw
bch82ZO24T7Or3lcJgN+5bo37evA/LrT4llwwo/zncZj3C8LN2cO/BQw3BL0
AvNr71fmg+rsnMTtHsyfyr9UCIkPk7RrXmH/opObikHB+R1l4j3mW28WLoVY
Fb+MdQNonnKtMvjMyUcTHsL7I3pkvgyuNpr9Nh/BehGDIxXQH73E4f8d43Ny
jytAyjpVMmkC63ds0K0EiozSqcopdHfJs0p4u2bQ4cUsXk+JZagCsQ8k6vff
2E/pc24VBJUKZrH+wedhQZ25GnrIbS27lrFejt2oGvYYXvigxLia2E+7eFo1
+OxjXjBlXo31cSt/75/9zd7qw0rYUvbAxhrY2aVxMGEdYQGOJdkacMv6qlO2
YTXON3u2Bjq8wh2fbSKsHLTJtga2aYlEjHCgpc5crgGn7Y+zmbmx38Y6hxqg
Tzq0CvDi+g81TWtWvn/W9StsJzz4foNqDdglFS0aC2D/TPYdNdBw6SS35y7M
Vzn3sxrY4Yd0nDDOW7RUXQ1Wm6N1S/YRTs9kd6uGqmGJy4/Fcf2KjF3VsKbu
ReSXA9hf/MWTKjC95pLLKI3eX2FXBffMN9/fLov9btgvVgLTwYoBOTnst7wl
tBIMVp/9a6CA63f3MlZC/tsZHjcl7C/W61UBf4sSZKJVMX9UbqgcTgfL6BVq
4HllSamXw9we72tD2lg/v+ddKWgtcOcv66Dr2mzvwd1nde28erhfO67PxaDu
vrikZ4zx92UFBRBHD3S5b4Pz2YUEZMLwjR3XB+zQh4Kvp8MRu5aCP5cwn61G
6C70b2D8LO2KttAcSYCDnzL+nfZAi9fmx0JYlSq/kzfaal1+BLwN/3wkwg89
fO5eMIibUgyyA9Eu96mOYOpas4U5BG18d60+XA0feXGBgpaUdPeEuru81++H
o7+ue0yGb5UntXZdQ/+8W3wVuJ4EsYZG47zZLZJxoD5Y1j4Uh44LmE6ATLat
SpnJmB9cLXwHuoSO/2W8g3FSa2Ia/DviW2+Vhn59xyMDzC5+lBbMxXy7G8nZ
cM1/4xSpAOMCp7bkQkOcyr3BYjzPC0f984CbliOSXone+m/le/tY75sv/2rQ
iutOFYHH2NosiwbsZ1DtXAxZDAqWLc0YX1scWAI9XM7bd7aiW9J97wHD/vS+
oPb/7/eT1qVwQK0nqf8h1vvX8pWB+TlmA3iK98MRu4AyiLoiuyW1E+//Ub6X
ZdAYZv9iqRvnPd6ysxzGUm5HnX+F/c6AdTnwVDzTbH6L652IvF0Oxx8ts2z/
gP3v3XhUDl79ku0BA5ivqD9eDtmz1iEfhnC97+2rK+Dlupug8BXjw0OcFcAk
+PBPyijej5R721Z+Xw4v1P0Zx3wjsZW45SkxL9NJ7L/NkrkCoi+cl26cwfwC
w7FyaPaNmeT9jfNv4u8oh/GY1hK/RZzPl55UDry5M5f6ltAHjCzKQbNpj4g8
IwuRf2uOvxx8eoy+3GImTJet6yqD3JHIzAVWwgxpdX5l8Gq50eLcesIkf+6V
82PeOsFfz45e//NCKRwUE+jj2YJesxR0D6xU9JJ8thJWjioML4FYI4r+Wx60
xumgYvgROtKZtBP7b/8jWwj8t3mjfgthPOXFQj5olZ3UNBJGM8oW50H+h7L7
XPuxf3LCp2ywlvGtuymPjnsudxfitIs8fymi5T/23gK61ceDBirofeVGibA9
WqWE8wSul5QlHQ0ns90veWijH2bUUsGvIWdvrw567rNYEBR0vRmW1kMr/NW4
DG+/rs28YYBWvrFzPzzSj7xTb4yejRJ2gLq2NUkDpug7h1l9IDmVNUrMGu3z
aUc4RLCFU89cRG97EBAFPn4sId72ON8t/Q+xYGS42uu+M/qd6NBNONZOcRl1
/X/9OyeSQfYgs+NGT/QekZTbwMW+ytLUH/OVVzGlAUsA2SQkCOOHYq+mw9wo
o0FeCF7PhYWIDHj9gEFrJhzzfxpfyIIOaZL6tmt4vhO9kA01Gf9AKRrz78mx
50DuxmC5i3HYz1a/OwcSA5elryVgfN9ceC5QvwdKlCehJ2ek8sDr3JLIm9vY
75rwizywfRiwa+ku1qdfssgHA5m/23dlYDwzfSgf1LP8eTSzsf5tsXEBHNr8
Z8uVPJxP27u1AHYH+224WYjxJ9/5C4HzxwJrYwn608LK33dmU1+mT2XY701C
USHMPpr/y1KF9QWV/YUwLOvzW7wWveoMcxH0Zv+e0mvAeZQc+YugfYv3uE8z
1p9n3lsEVaS5r6l07P9PcHcRZE94fmq/j3Z8zlEECWa/3o91YL0fx1whUJ54
vN70BOtTmZ8WgseR2S7Z57jem6r4QrDJdX9q1oWuOnqqEM5yzjwgv0RrVi0W
gCrZjZ7/Gr1fMaUADk5ONXS+w/5u8xIFIGTuWj37AdcvWl+TD1ueTZbyDuJ6
79Ik84FJ3qVQ+TPmO39KzYPpvJ/Ztl/R2gyMedATOnGrYhz7n8rNyIG2qcsJ
b39iv52pg9lQYfEjenkaHaHPmQ3xR8dDtRawv+Qq00wgFzgGufzFfjSqcwa4
cX/3SfyHcdcB73Q4MzN6eWg1KxFvvNl3B5StHOzXrEV/XF11GyRfjFhLsBEm
ZVb7J8Omom9GflvQT0vf3oAu6y8qW3ZiftqnplD4zJw3f1MIzfuL7g9zufal
24TRH8rTrgDf9+/8gvvRTZOnrGi27tO/DxxF257Jv07z2Vp1rxLQr7M2xNOu
1XpePKyKXtqmdZNW9me+BzTRHC0PUmh/Qv+VnDIirMx99FQ2jV241abbhDCd
Q3MghybwiMxnYI75i6JOebRjbCyRZhcxP7bdrZBmfO+h0qA95n9IHSuiOepG
zl1wwv1ZxJ8rocUmsNk4umN+ijNbGS1LtpP3pxfmH0upLKPVvI3pdvMjnJ5k
qF9Oe+R3JmIukLAAD/9EOe09P4eSXwjGTS6SKmgTtN5fSxTClqsK11XSGKwS
i0MisN9Di6hKGscq4wuro3B9j0XmKppwzjbeyBjMn/zuXkWTO/6+i+0GWj37
bRXt5Oid8NhErL8dd7CaZn7NHDhvo9UFQqpprvsFfiXdJTxIb7hfTaO8+FTE
l4HzND1eqKYlumZZp2Xj/rZVCdXQCjgubtuVj/Gf/Yo1tKZq4a7cIownlmjV
0F4YjVBFS7F/RsiJGtrQQoHivQrMH2+QraHNpjjOStWgk2s5a2issL+ouh7P
g7P3UzWNd3DCSq4ZLeB4t5omTi7jaaaju14dq6Yp7XZ7odyO88Y6fqyi6XVI
U9sf4vr7yZZVtIv2vxROPMX9d19+UUnzWVc787QTz4stZF8l7WqxT+HpHoz3
sTlX0O7qyFu9fIX9MyzvlNPKJv9wG73D+0Hhycr1a4tv7uz7gPXTYnaltFeH
gsPMB3E9te2yJbSR18oKQ5/Ri4s9hbRFH6aZi9/QRfoK+bQNfO0Fo2N4HrM1
Liv3W3OY5eUJ7B9d5JxJO2hxnHtqCvtfd3NIpWkwru30+IWWekROohlnPaHM
z6NPmwjH0i5pRB0N+Iv1atqZJFrgt1PT//6hV37R05v/Aw6mnvM=
     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.5086608359503593`*^9, 3.508660885434101*^9}]
}, Open  ]],

Cell["\<\
Find difference between our approximation and Mathematicas approximation.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData["SetPrecision"], "Input"],

Cell[BoxData["SetPrecision"], "Output",
 CellChangeTimes->{3.5086608360284834`*^9, 3.5086608854966*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 StyleBox[
  RowBox[{"$MaxPrecision", "=", "100"}],
  FontColor->RGBColor[1, 0, 1]]], "Input"],

Cell[BoxData["100"], "Output",
 CellChangeTimes->{3.508660836044108*^9, 3.508660885512225*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"SetPrecision", "[", 
  RowBox[{
   RowBox[{"Plot", "[", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{
       RowBox[{"g", "[", "x", "]"}], "-", 
       RowBox[{"InterpolatingFunction1D", "[", "x", "]"}]}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"x", ",", 
       RowBox[{"-", "2."}], ",", "3."}], "}"}], ",", 
     RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}], ",", "100"}], 
  "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], 
    LineBox[{{-1.9999998979591838`, 0}, {-1.998466410397202, 
       0}, {-1.9969329228352202`, 
      8.881784197001252*^-16}, {-1.9953994352732383`, 
       0}, {-1.9938659477112566`, 
       0}, {-1.992332460149275, -8.881784197001252*^-16}, \
{-1.990798972587293, -8.881784197001252*^-16}, {-1.989265485025311, 
      8.881784197001252*^-16}, {-1.9877319974633294`, 
      8.881784197001252*^-16}, {-1.9861985099013477`, 
       0}, {-1.9846650223393658`, 0}, {-1.983131534777384, 
      8.881784197001252*^-16}, {-1.9815980472154022`, \
-8.881784197001252*^-16}, {-1.9800645596534203`, 0}, {-1.9785310720914384`, 
       0}, {-1.9769975845294567`, 0}, {-1.9754640969674748`, 
       0}, {-1.950928295975766, 0}, {-1.9018566939923482`, 
       0}, {-1.9001942177514164`, 0}, {-1.8985317415104843`, 
      8.881784197001252*^-16}, {-1.8968692652695525`, 
       0}, {-1.8952067890286206`, 0}, {-1.8935443127876888`, 
       0}, {-1.891881836546757, 
      8.881784197001252*^-16}, {-1.890219360305825, 
       0}, {-1.8885568840648932`, -8.881784197001252*^-16}, \
{-1.8868944078239613`, 0}, {-1.8852319315830295`, 0}, {-1.8835694553420974`, 
       0}, {-1.8819069791011656`, 0}, {-1.8802445028602337`, 
       0}, {-1.8785820266193016`, 0}, {-1.875257074137438, 
       0}, {-1.8486574542825276`, 0}, {-1.8469949780415957`, 
       0}, {-1.8453325018006637`, -8.881784197001252*^-16}, \
{-1.8436700255597318`, 0}, {-1.8420075493188, 0}, {-1.8353576443550725`, 
       0}, {-1.8336951681141407`, 0}, {-1.8320326918732088`, 
       0}, {-1.8303702156322768`, 
       0}, {-1.828707739391345, -8.881784197001252*^-16}, \
{-1.827045263150413, -8.881784197001252*^-16}, {-1.825382786909481, 
       0}, {-1.8237203106685491`, 0}, {-1.8220578344276173`, 
      8.881784197001252*^-16}, {-1.8203953581866854`, 
       0}, {-1.8187328819457536`, 
       0}, {-1.8170704057048215`, -8.881784197001252*^-16}, \
{-1.8154079294638896`, 0}, {-1.8137454532229578`, 0}, {-1.8120829769820257`, 
       0}, {-1.8104205007410938`, -8.881784197001252*^-16}, \
{-1.808758024500162, -8.881784197001252*^-16}, {-1.8070955482592301`, \
-8.881784197001252*^-16}, {-1.8054330720182983`, 
      8.881784197001252*^-16}, {-1.8037705957773664`, 
       0}, {-1.8021081195364346`, 
      8.881784197001252*^-16}, {-1.8004456432955027`, 
       0}, {-1.7987831670545709`, 0}, {-1.7971206908136388`, 
       0}, {-1.795458214572707, 
      8.881784197001252*^-16}, {-1.7923536010594736`, 
      8.881784197001252*^-16}, {-1.78924898754624, 
      8.881784197001252*^-16}, {-1.7876966807896233`, 
       0}, {-1.7861443740330067`, 
      8.881784197001252*^-16}, {-1.7845920672763902`, 
       0}, {-1.7830397605197734`, 
       0}, {-1.7814874537631566`, -8.881784197001252*^-16}, \
{-1.77993514700654, 
      8.881784197001252*^-16}, {-1.7783828402499235`, 
       0}, {-1.7768305334933068`, 8.881784197001252*^-16}, {-1.77527822673669,
       8.881784197001252*^-16}, {-1.7737259199800735`, 
       0}, {-1.772173613223457, 0}, {-1.7706213064668401`, 
       0}, {-1.7690689997102234`, -8.881784197001252*^-16}, \
{-1.7675166929536068`, 0}, {-1.7659643861969903`, 
       0}, {-1.7644120794403735`, -8.881784197001252*^-16}, \
{-1.76130746592714, -8.881784197001252*^-16}, {-1.7582028524139066`, \
-8.881784197001252*^-16}, {-1.75665054565729, 0}, {-1.7550982389006733`, 
       0}, {-1.7535459321440565`, 0}, {-1.7519936253874397`, 
       0}, {-1.750441318630823, -8.881784197001252*^-16}, \
{-1.7488890118742064`, 0}, {-1.745784398360973, 0}, {-1.7442320916043563`, 
       0}, {-1.7426797848477398`, 
      8.881784197001252*^-16}, {-1.7411274780911232`, 
       0}, {-1.7395751713345065`, 
      8.881784197001252*^-16}, {-1.7380228645778897`, 
       0}, {-1.7364705578212731`, 0}, {-1.7349182510646566`, 
       0}, {-1.7333659443080398`, 
      8.881784197001252*^-16}, {-1.731813637551423, 0}, {-1.7302613307948065`,
        0}, {-1.72870902403819, 
       0}, {-1.7271567172815732`, -8.881784197001252*^-16}, \
{-1.7256044105249564`, 
       0}, {-1.7240521037683396`, -8.881784197001252*^-16}, \
{-1.7224997970117228`, 0}, {-1.7209474902551063`, 0}, {-1.7193951834984897`, 
       0}, {-1.717842876741873, 
      8.881784197001252*^-16}, {-1.7162905699852562`, 
       0}, {-1.7147382632286394`, 
      8.881784197001252*^-16}, {-1.7131859564720227`, 
      8.881784197001252*^-16}, {-1.711633649715406, -8.881784197001252*^-16}, \
{-1.7100813429587896`, 0}, {-1.7085290362021728`, 
      8.881784197001252*^-16}, {-1.706976729445556, 0}, {-1.7054244226889395`,
       8.881784197001252*^-16}, {-1.703872115932323, 
       0}, {-1.7023198091757061`, -8.881784197001252*^-16}, \
{-1.7007675024190894`, 0}, {-1.6992151956624728`, 0}, {-1.6976628889058563`, 
       0}, {-1.6961105821492395`, 
      8.881784197001252*^-16}, {-1.694588725489183, -8.881784197001252*^-16}, \
{-1.6930668688291264`, 
      8.881784197001252*^-16}, {-1.6915450121690698`, 
       0}, {-1.6900231555090133`, 0}, {-1.6885012988489567`, 
       0}, {-1.6869794421889002`, 0}, {-1.683935728868787, 
       0}, {-1.6824138722087305`, 0}, {-1.680892015548674, 
       0}, {-1.6793701588886174`, 0}, {-1.6778483022285609`, 
      8.881784197001252*^-16}, {-1.6763264455685043`, 
       0}, {-1.6748045889084477`, 0}, {-1.6732827322483912`, 
       0}, {-1.6717608755883346`, 0}, {-1.670239018928278, 
       0}, {-1.6687171622682215`, 0}, {-1.6656734489481084`, 
       0}, {-1.6641515922880519`, 
      8.881784197001252*^-16}, {-1.6626297356279953`, 
       0}, {-1.6611078789679388`, 0}, {-1.6595860223078822`, 
      8.881784197001252*^-16}, {-1.6580641656478257`, 
       0}, {-1.656542308987769, -8.881784197001252*^-16}, \
{-1.6550204523277126`, -8.881784197001252*^-16}, {-1.653498595667656, \
-8.881784197001252*^-16}, {-1.6519767390075994`, 
      8.881784197001252*^-16}, {-1.650454882347543, 0}, {-1.6489330256874863`,
        0}, {-1.6474111690274298`, 0}, {-1.6458893123673732`, 
       0}, {-1.6443674557073167`, 
      8.881784197001252*^-16}, {-1.6428455990472601`, 
       0}, {-1.6413237423872036`, 0}, {-1.639801885727147, 
      8.881784197001252*^-16}, {-1.6382800290670905`, 0}, {-1.636758172407034,
        0}, {-1.6352363157469774`, -8.881784197001252*^-16}, \
{-1.6337144590869208`, -8.881784197001252*^-16}, {-1.6321926024268643`, 
       0}, {-1.6306707457668077`, 
      8.881784197001252*^-16}, {-1.6291488891067512`, \
-8.881784197001252*^-16}, {-1.6276270324466946`, -8.881784197001252*^-16}, \
{-1.626105175786638, 0}, {-1.6245833191265815`, 0}, {-1.623061462466525, 
      8.881784197001252*^-16}, {-1.6215396058064684`, \
-8.881784197001252*^-16}, {-1.6200177491464118`, 0}, {-1.6184958924863553`, 
       0}, {-1.6169740358262987`, 
      8.881784197001252*^-16}, {-1.6154521791662422`, 
       0}, {-1.6139303225061856`, -8.881784197001252*^-16}, \
{-1.612408465846129, 
      8.881784197001252*^-16}, {-1.6108866091860725`, 0}, {-1.609364752526016,
        0}, {-1.6078428958659594`, -8.881784197001252*^-16}, \
{-1.6063210392059029`, -8.881784197001252*^-16}, {-1.6047991825458463`, 
       0}, {-1.5987117559056199`, 0}, {-1.5970609105666131`, 
      8.881784197001252*^-16}, {-1.5954100652276066`, 
       0}, {-1.5921083745495932`, 0}, {-1.5855049931935667`, 
       0}, {-1.5722982304815134`, 0}, {-1.570647385142507, 
       0}, {-1.5689965398035002`, 0}, {-1.5673456944644935`, 
      8.881784197001252*^-16}, {-1.565694849125487, -8.881784197001252*^-16}, \
{-1.5640440037864802`, 0}, {-1.5623931584474735`, 
       0}, {-1.560742313108467, -8.881784197001252*^-16}, \
{-1.5590914677694603`, -8.881784197001252*^-16}, {-1.5574406224304536`, 
      8.881784197001252*^-16}, {-1.555789777091447, 0}, {-1.5541389317524403`,
        0}, {-1.5524880864134336`, 
       0}, {-1.550837241074427, -8.881784197001252*^-16}, \
{-1.5491863957354204`, 0}, {-1.5475355503964137`, 0}, {-1.5458847050574072`, 
      8.881784197001252*^-16}, {-1.5442338597184007`, 0}, {-1.542583014379394,
       8.881784197001252*^-16}, {-1.5409321690403872`, 
       0}, {-1.5392813237013807`, 0}, {-1.537630478362374, 
       0}, {-1.5359796330233673`, 0}, {-1.532677942345354, 
       0}, {-1.5310270970063473`, 
      8.881784197001252*^-16}, {-1.5293762516673408`, 
       0}, {-1.5260745609893274`, 0}, {-1.519471179633301, 
       0}, {-1.5178203342942942`, 
      8.881784197001252*^-16}, {-1.5161694889552875`, 
       0}, {-1.5128677982772742`, 0}, {-1.5062644169212476`, 
       0}, {-1.4930576542091945`, 0}, {-1.491516978354503, 
       0}, {-1.4899763024998114`, 
       0}, {-1.48843562664512, -8.881784197001252*^-16}, \
{-1.4868949507904285`, 
      8.881784197001252*^-16}, {-1.4853542749357371`, 
       0}, {-1.4838135990810457`, 
      8.881784197001252*^-16}, {-1.4822729232263543`, 
       0}, {-1.4807322473716629`, 0}, {-1.4791915715169714`, 
       0}, {-1.47765089566228, 0}, {-1.4745695439528972`, 
       0}, {-1.4730288680982058`, 0}, {-1.4714881922435143`, 
       0}, {-1.469947516388823, 0}, {-1.4684068405341315`, 
      8.881784197001252*^-16}, {-1.46686616467944, 0}, {-1.4653254888247487`, 
       0}, {-1.4637848129700572`, 0}, {-1.4622441371153658`, 
       0}, {-1.4607034612606744`, 0}, {-1.459162785405983, 
       0}, {-1.4560814336966001`, 0}, {-1.4545407578419087`, 
       0}, {-1.4530000819872173`, 0}, {-1.4514594061325259`, 
      8.881784197001252*^-16}, {-1.4499187302778345`, 
      8.881784197001252*^-16}, {-1.448378054423143, 0}, {-1.4468373785684516`,
        0}, {-1.4452967027137602`, 0}, {-1.4437560268590688`, 
       0}, {-1.4422153510043774`, 0}, {-1.440674675149686, 
       0}, {-1.437593323440303, 0}, {-1.4314306200215374`, 
       0}, {-1.4191052131840058`, 0}, {-1.4175645373293144`, 
       0}, {-1.4160238614746228`, 
      4.440892098500626*^-16}, {-1.4144831856199314`, \
-4.440892098500626*^-16}, {-1.41294250976524, -4.440892098500626*^-16}, \
{-1.4114018339105485`, 
       0}, {-1.409861158055857, -4.440892098500626*^-16}, \
{-1.4083204822011657`, 0}, {-1.4067798063464743`, 
       0}, {-1.4052391304917828`, -4.440892098500626*^-16}, \
{-1.4036984546370914`, 0}, {-1.4021577787824, 
      4.440892098500626*^-16}, {-1.4006171029277086`, \
-4.440892098500626*^-16}, {-1.3990764270730172`, 
       0}, {-1.3975357512183257`, -4.440892098500626*^-16}, \
{-1.3959950753636343`, 0}, {-1.394454399508943, 
      4.440892098500626*^-16}, {-1.3927847349753013`, 0}, {-1.39111507044166, 
       0}, {-1.3894454059080186`, -4.440892098500626*^-16}, \
{-1.387775741374377, -4.440892098500626*^-16}, {-1.3861060768407354`, 
      4.440892098500626*^-16}, {-1.3844364123070938`, 
       0}, {-1.3827667477734522`, 0}, {-1.3810970832398108`, 
       0}, {-1.3794274187061695`, 0}, {-1.3777577541725279`, 
       0}, {-1.3744184251052447`, 0}, {-1.3677397669706788`, 
       0}, {-1.3660701024370372`, 0}, {-1.3644004379033956`, 
      4.440892098500626*^-16}, {-1.362730773369754, 
       0}, {-1.3610611088361126`, -4.440892098500626*^-16}, \
{-1.3593914443024713`, 0}, {-1.3577217797688297`, 0}, {-1.356052115235188, 
      4.440892098500626*^-16}, {-1.3543824507015465`, 0}, {-1.352712786167905,
        0}, {-1.3510431216342635`, 
      4.440892098500626*^-16}, {-1.3493734571006222`, 
       0}, {-1.3477037925669806`, 0}, {-1.346034128033339, 
       0}, {-1.3443644634996974`, 0}, {-1.3426947989660558`, 
      4.440892098500626*^-16}, {-1.3410251344324144`, 
      4.440892098500626*^-16}, {-1.339355469898773, 0}, {-1.3376858053651315`,
        0}, {-1.33601614083149, 
      4.440892098500626*^-16}, {-1.3343464762978483`, 
       0}, {-1.3326768117642067`, 
      4.440892098500626*^-16}, {-1.3310071472305653`, 
       0}, {-1.3276678181632824`, 
       0}, {-1.3259981536296408`, -4.440892098500626*^-16}, \
{-1.3243284890959992`, 0}, {-1.3209891600287162`, 0}, {-1.31431050189415, 
       0}, {-1.3126408373605085`, 0}, {-1.3109711728268671`, 
       0}, {-1.3076318437595842`, 0}, {-1.3059621792259426`, 
      4.440892098500626*^-16}, {-1.304292514692301, 0}, {-1.3026228501586594`,
        0}, {-1.300953185625018, -4.440892098500626*^-16}, \
{-1.2992835210913767`, 0}, {-1.297613856557735, 
       0}, {-1.2959441920240935`, -4.440892098500626*^-16}, \
{-1.294274527490452, 0}, {-1.2926048629568103`, 0}, {-1.290935198423169, 
       0}, {-1.287595869355886, 0}, {-1.2859566549188046`, 
       0}, {-1.2843174404817232`, 0}, {-1.2810390116075605`, 
       0}, {-1.2744821538592348`, 0}, {-1.2613684383625836`, 
       0}, {-1.2351410073692812`, 
       0}, {-1.2335017929321999`, -4.440892098500626*^-16}, \
{-1.2318625784951185`, 0}, {-1.2302233640580371`, 0}, {-1.2285841496209557`, 
      4.440892098500626*^-16}, {-1.2269449351838744`, 0}, {-1.225305720746793,
        0}, {-1.2236665063097116`, 
       0}, {-1.22202729187263, -4.440892098500626*^-16}, \
{-1.2203880774355484`, 0}, {-1.218748862998467, 
      4.440892098500626*^-16}, {-1.2171096485613857`, 
       0}, {-1.2154704341243043`, 
      4.440892098500626*^-16}, {-1.213831219687223, 
      4.440892098500626*^-16}, {-1.2121920052501416`, \
-4.440892098500626*^-16}, {-1.2105527908130602`, -4.440892098500626*^-16}, \
{-1.2089135763759788`, 0}, {-1.2072743619388975`, 0}, {-1.205635147501816, 
       0}, {-1.2023567186276534`, 0}, {-1.1957998608793279`, 
       0}, {-1.1941606464422465`, 0}, {-1.1925214320051651`, 
       0}, {-1.1908822175680838`, 0}, {-1.1892430031310024`, 
      4.440892098500626*^-16}, {-1.187603788693921, 0}, {-1.1859645742568397`,
        0}, {-1.1843253598197583`, -4.440892098500626*^-16}, \
{-1.1826861453826767`, -4.440892098500626*^-16}, {-1.1337567068941588`, \
-4.440892098500626*^-16}, {-1.084827268405641, -4.440892098500626*^-16}, \
{-0.9787131159758, -4.440892098500626*^-16}, {-0.977165251828399, 
       0}, {-0.9756173876809979, 0}, {-0.9740695235335969, 
       0}, {-0.9725216593861958, 
      4.440892098500626*^-16}, {-0.9709737952387948, 
      4.440892098500626*^-16}, {-0.9694259310913937, 
       0}, {-0.9678780669439927, -4.440892098500626*^-16}, \
{-0.9663302027965917, 0}, {-0.9647823386491906, 
       0}, {-0.9632344745017896, -4.440892098500626*^-16}, \
{-0.9616866103543885, -4.440892098500626*^-16}, {-0.9601387462069875, 
       0}, {-0.9585908820595864, 0}, {-0.9570430179121854, 
       0}, {-0.9554951537647843, 
       0}, {-0.9539472896173833, -4.440892098500626*^-16}, \
{-0.9291814632589666, -2.220446049250313*^-16}, {-0.9167985500797582, \
-2.220446049250313*^-16}, {-0.9044156369005498, -2.220446049250313*^-16}, \
{-0.9028677727531488, 0}, {-0.9013199086057477, 
      2.220446049250313*^-16}, {-0.8997720444583467, 0}, {-0.8982241803109456,
        0}, {-0.8966763161635446, 0}, {-0.8951284520161435, 
       0}, {-0.8920327237213415, 
      2.220446049250313*^-16}, {-0.8904848595739404, 0}, {-0.8889369954265394,
        0}, {-0.8858412671317373, 0}, {-0.879649810542133, 
      2.220446049250313*^-16}, {-0.8781323964912922, 0}, {-0.8766149824404512,
        0}, {-0.8735801543387693, 0}, {-0.8675104981354056, 
       0}, {-0.8553711857286781, 0}, {-0.8310925609152232, 
       0}, {-0.8295751468643824, 
       0}, {-0.8280577328135414, -2.220446049250313*^-16}, \
{-0.8250229047118596, 
       0}, {-0.8189532485084958, -2.220446049250313*^-16}, \
{-0.8068139361017683, 
       0}, {-0.7825353112883134, -2.220446049250313*^-16}, \
{-0.7808889085585224, 
      2.220446049250313*^-16}, {-0.7792425058287313, 0}, {-0.7759497003691493,
        0}, {-0.7693640894499854, -2.220446049250313*^-16}, \
{-0.7561928676116572, 0}, {-0.7298504239350009, 0}, {-0.6771655365816883, 
       0}, {-0.5788466088712372, 0}, {-0.47227240570798046`, 
       0}, {-0.3676470087245712, 0}, {-0.27007245873733593`, 
       0}, {-0.1642426332972951, 
       0}, {-0.06546365485342827, -1.3877787807814457`*^-17}, {
      0.041570599043244144`, 0}, {0.14665604676006908`, 0}, {
      0.24469064748072, 0}, {0.35098052365417653`, 0}, {0.45021955283145904`, 
      5.551115123125783*^-17}, {
      0.5475097758288942, -5.551115123125783*^-17}, {0.6530552742791348, 0}, {
      0.7515499257332015, 0}, {0.8582998526400737, 
      2.7755575615628914`*^-17}, {
      0.9631009733670984, -2.7755575615628914`*^-17}, {
      1.0608512470979492`, 0}, {1.1668567962816054`, 0}, {
      1.2658114984690878`, 0}, {
      1.3628173944767228`, -1.3877787807814457`*^-17}, {
      1.4680785659371633`, 0}, {1.5662888904014298`, 0}, {1.6727544903185019`,
       5.551115123125783*^-17}, {1.7721692432394, 0}, {
      1.8696351899804504`, 0}, {1.9753564121743066`, 0}, {
      2.074026787371989, 0}, {2.1809524380224765`, 0}, {
      2.2859292824931168`, 0}, {2.383855279967583, 0}, {
      2.4900365528948547`, 0}, {
      2.4915854658000285`, -2.220446049250313*^-16}, {2.493134378705202, 0}, {
      2.4962322045155485`, -2.220446049250313*^-16}, {
      2.502427856136242, -2.220446049250313*^-16}, {2.5148191593776295`, 0}, {
      2.5396017658604038`, 0}, {2.541150678765577, 0}, {
      2.5426995916707504`, 0}, {2.545797417481097, 0}, {2.551993069101791, 
      2.220446049250313*^-16}, {2.564384372343178, 2.220446049250313*^-16}, {
      2.5659332852483514`, -2.220446049250313*^-16}, {
      2.5674821981535247`, -4.440892098500626*^-16}, {
      2.569031111058698, -4.440892098500626*^-16}, {
      2.5705800239638714`, -4.440892098500626*^-16}, {
      2.5721289368690448`, 0}, {2.573677849774218, 4.440892098500626*^-16}, {
      2.5752267626793914`, -4.440892098500626*^-16}, {
      2.576775675584565, -4.440892098500626*^-16}, {2.578324588489739, 0}, {
      2.5798735013949123`, -4.440892098500626*^-16}, {
      2.5814224143000857`, 0}, {2.582971327205259, 4.440892098500626*^-16}, {
      2.5845202401104324`, 0}, {2.5860691530156057`, 0}, {
      2.587618065920779, 0}, {2.5891669788259524`, 4.440892098500626*^-16}, {
      2.590685441634566, 0}, {2.592203904443179, 4.440892098500626*^-16}, {
      2.593722367251792, 4.440892098500626*^-16}, {2.5952408300604057`, 0}, {
      2.596759292869019, -4.440892098500626*^-16}, {2.598277755677632, 0}, {
      2.6013146812948587`, 0}, {2.6028331441034718`, 
      4.440892098500626*^-16}, {2.6043516069120853`, 0}, {
      2.6073885325293116`, 0}, {2.613462383763765, 0}, {
      2.6149808465723785`, -4.440892098500626*^-16}, {
      2.6164993093809916`, 0}, {2.6180177721896047`, 0}, {
      2.6195362349982183`, -4.440892098500626*^-16}, {
      2.6210546978068314`, 0}, {
      2.6225731606154445`, -4.440892098500626*^-16}, {
      2.624091623424058, -4.440892098500626*^-16}, {2.6256100862326712`, 
      4.440892098500626*^-16}, {2.6271285490412843`, 0}, {
      2.628647011849898, 0}, {2.630165474658511, 4.440892098500626*^-16}, {
      2.631683937467124, 0}, {2.6332024002757377`, 0}, {
      2.634720863084351, 0}, {2.636239325892964, 0}, {
      2.6377577887015775`, -4.440892098500626*^-16}, {
      2.649905491170484, -4.440892098500626*^-16}, {
      2.66205319363939, -4.440892098500626*^-16}, {2.6635716564480036`, 0}, {
      2.6650901192566168`, 4.440892098500626*^-16}, {2.66660858206523, 0}, {
      2.6681270448738434`, 0}, {2.6696455076824566`, 
      4.440892098500626*^-16}, {2.6711639704910697`, 0}, {
      2.6742008961082964`, 0}, {2.6757193589169095`, 0}, {
      2.677237821725523, 0}, {2.6802747473427493`, 0}, {
      2.6863485985772027`, 0}, {2.687996050064766, -4.440892098500626*^-16}, {
      2.6896435015523297`, 4.440892098500626*^-16}, {2.6912909530398927`, 
      4.440892098500626*^-16}, {2.6929384045274563`, 0}, {
      2.69952821047771, 0}, {2.701175661965273, 0}, {
      2.7028231134528364`, -4.440892098500626*^-16}, {2.7044705649404, 
      4.440892098500626*^-16}, {2.706118016427963, 0}, {
      2.7077654679155265`, 0}, {2.70941291940309, 0}, {
      2.711060370890653, 0}, {2.7127078223782166`, -4.440892098500626*^-16}, {
      2.71435527386578, 0}, {2.716002725353343, 4.440892098500626*^-16}, {
      2.7176501768409067`, 0}, {2.71929762832847, 4.440892098500626*^-16}, {
      2.7209450798160333`, 0}, {2.722592531303597, 0}, {
      2.7242399827911603`, 0}, {
      2.7258874342787234`, -4.440892098500626*^-16}, {2.727534885766287, 0}, {
      2.7291823372538504`, 0}, {2.7308297887414135`, 
      4.440892098500626*^-16}, {2.732477240228977, 0}, {
      2.7341246917165405`, 0}, {2.7357721432041036`, 0}, {2.737419594691667, 
      4.440892098500626*^-16}, {2.7390670461792306`, 
      4.440892098500626*^-16}, {2.740714497666794, -4.440892098500626*^-16}, {
      2.7423619491543576`, 0}, {
      2.7440094006419207`, -4.440892098500626*^-16}, {2.745656852129484, 0}, {
      2.7473043036170477`, 0}, {2.7489517551046108`, 
      4.440892098500626*^-16}, {2.7505992065921743`, 
      4.440892098500626*^-16}, {2.752246658079738, 0}, {
      2.753894109567301, -4.440892098500626*^-16}, {2.7555415610548644`, 0}, {
      2.758836464029991, 0}, {2.7654262699802445`, 0}, {
      2.767073721467808, -4.440892098500626*^-16}, {
      2.768721172955371, -4.440892098500626*^-16}, {2.7703686244429346`, 0}, {
      2.772016075930498, -4.440892098500626*^-16}, {2.773663527418061, 0}, {
      2.7753109789056247`, 0}, {2.7769584303931882`, 0}, {
      2.7786058818807513`, 0}, {2.7917854937812585`, 0}, {
      2.8901715419891403`, 0}, {2.8918876100511723`, 0}, {
      2.8936036781132044`, 0}, {2.895319746175236, -4.440892098500626*^-16}, {
      2.897035814237268, -4.440892098500626*^-16}, {2.8987518822993, 
      4.440892098500626*^-16}, {2.9004679503613318`, 0}, {2.902184018423364, 
      4.440892098500626*^-16}, {2.903900086485396, 0}, {
      2.917628630981651, 0}, {2.919344699043683, 0}, {
      2.9210607671057147`, 0}, {2.922776835167747, 0}, {
      2.924492903229779, -4.440892098500626*^-16}, {2.9262089712918105`, 0}, {
      2.9279250393538425`, 0}, {2.9296411074158746`, 0}, {
      2.931357175477906, 0}, {2.9330732435399383`, -4.440892098500626*^-16}, {
      2.9347893116019703`, 0}, {2.938221447726034, 0}, {
      2.939937515788066, 0}, {2.9416535838500977`, 0}, {
      2.9433696519121297`, -4.440892098500626*^-16}, {
      2.945085719974162, -4.440892098500626*^-16}, {2.946801788036194, 0}, {
      2.948517856098226, 0}, {2.9502339241602575`, 0}, {
      2.9519499922222896`, -4.440892098500626*^-16}, {
      2.9536660602843217`, 0}, {
      2.9553821283463533`, -4.440892098500626*^-16}, {2.9570981964083853`, 
      4.440892098500626*^-16}, {2.9588142644704174`, 0}, {
      2.9605303325324495`, 0}, {2.9622464005944815`, 
      4.440892098500626*^-16}, {2.963962468656513, 0}, {
      2.965678536718545, 0}, {2.972542808966673, 0}, {2.974258877028705, 0}, {
      2.975974945090737, 0}, {2.9776910131527687`, 0}, {
      2.979407081214801, -4.440892098500626*^-16}, {2.981123149276833, 0}, {
      2.9828392173388645`, -4.440892098500626*^-16}, {
      2.9845552854008965`, 0}, {2.9862713534629286`, 0}, {2.98798742152496, 
      4.440892098500626*^-16}, {2.9897034895869923`, 0}, {
      2.993135625711056, 0}, {2.994851693773088, 0}, {2.99656776183512, 0}, {
      2.9982838298971517`, 0}, {2.9999998979591838`, 0}}]}},
  AspectRatio->0.6180339887498949,
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.508660836965971*^9, 3.508660886418463*^9}]
}, Open  ]],

Cell["Good quality ", "Text"],

Cell["Install compiled code again", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"InstallCode", "[", "]"}], ";"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660837044095*^9, 3.508660886512212*^9}]
}, Open  ]],

Cell[TextData[{
 "Same comparison, but now use our compiled ",
 StyleBox["InterpolatingFunction1D",
  FontWeight->"Bold"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Plot", "[", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"1", ",", 
     RowBox[{"g", "[", "x", "]"}], ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"g", "[", "x", "]"}], "-", 
        RowBox[{"f", "[", "x", "]"}]}], ")"}], "*", 
      RowBox[{"10", "^", "6"}]}], ",", 
     RowBox[{
      RowBox[{"(", 
       RowBox[{
        RowBox[{"g", "[", "x", "]"}], "-", 
        RowBox[{"InterpolatingFunction1D", "[", "x", "]"}]}], ")"}], "*", 
      RowBox[{"10", "^", "6"}]}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"x", ",", 
     RowBox[{"-", "2."}], ",", "3"}], "}"}], ",", 
   RowBox[{"PlotRange", "\[Rule]", "All"}]}], "]"}]], "Input"],

Cell[BoxData[
 GraphicsBox[{{}, {}, 
   {Hue[0.67, 0.6, 0.6], LineBox[CompressedData["
1:eJxTTMoPSmViYGAwAWIQ/evc7Cf////fzwAGH+w1Xh6YxmCO4IczP/NgzPsH
53Of6ksx2vUHzlfb0H3TRfk3nL+0zfR40JSfcH71e9ucCY+/w/k7ymdlrff/
Buf7TQpV+DfzC5zvPnevn8ebT3D+suOV14qjP8L5ezg0tecFvofzfx9gixFX
ewPn/4+0q2vnegnnbxY35g9f/RTONxO5EPSy9RGcfykrhGub5T04/2DensD2
9utwvv+Hkn93PS7C+VVl/oLPWY/C+ar8ax5YHtoA54cE7b6212OpPYy/UtdH
bMHRQ3B+Q7fHG9ag83C+j/YK5vKya3C+5eTbq1Mv3IHzz89Za2Hc+hDOD/Te
oHPl6RM4n2U3V94mnhdwPjvbCgfD8tdwvmvBa8WdF9/B+bd8nTXsf35A2Bdo
xFG0+hOcb2h6tu+wwxc4/7jG7jmS577C+axdXp/ta7/D+bdPt35o4f8J50dW
J5nNOvwLzi/gfq10LuoPnK+8+tg0tld/4fz6Hc7B/LP/w/lWwWbRs6czOMD4
HI3ShtOLGOH8pWarvKZ6MsH5fXf6tisJMcP53pZHPGa8RvBD2vx1525jgfN1
oltCl9Sywvk3nnEumhzKBudfSHrqKCTPDuerLZWU6uxH8Guf+376/x/BBwCE
1w7u
     "]]}, 
   {Hue[0.9060679774997897, 0.6, 0.6], LineBox[CompressedData["
1:eJwVlnc81d8fx+1rVZJSaGigrxWhZLw/lJGMrAhRSlOUkRCVXUmR7JmoEBIl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     "]]}, 
   {Hue[0.14213595499957954`, 0.6, 0.6], LineBox[CompressedData["
1:eJwU2nlcTVsUB/Bb3VsqJVGKUsoYkpCQ1pGhopSEJE+DBkqSKBJXqNAcCUmS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     "]]}, 
   {Hue[0.37820393249936934`, 0.6, 0.6], LineBox[CompressedData["
1:eJxN23c41X//B3AkO5WUWRklkZlRcqwGGkiSJCvJiooUokJIZZa9s7Kzso7I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     "]]}},
  AspectRatio->NCache[GoldenRatio^(-1), 0.6180339887498948],
  Axes->True,
  AxesOrigin->{0, 0},
  PlotRange->{All, All},
  PlotRangeClipping->True,
  PlotRangePadding->{Automatic, Automatic}]], "Output",
 CellChangeTimes->{3.508660838903446*^9, 3.5086608867153344`*^9}]
}, Open  ]],

Cell["Result is same", "Text"],

Cell["Uninstall code", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"UninstallCode", "[", "]"}], ";"}], 
  "\[IndentingNewLine]"}]], "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["C++ code", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

#include \"Global.icc\"

#include <math.h>
void Global_TGlobalInit ()
{
    Global_Tx1a.SetShape(64 );
    Global_Tx1a(1) = -2.;
    Global_Tx1a(2) = -1.8958333333333335;
    Global_Tx1a(3) = -1.7916666666666665;
    Global_Tx1a(4) = -1.6875;
    Global_Tx1a(5) = -1.5833333333333335;
    Global_Tx1a(6) = -1.479166666666667;
    Global_Tx1a(7) = -1.375;
    Global_Tx1a(8) = -1.2708333333333333;
    Global_Tx1a(9) = -1.1666666666666667;
    Global_Tx1a(10) = -1.0625;
    Global_Tx1a(11) = -0.9583333333333334;
    Global_Tx1a(12) = -0.8541666666666667;
    Global_Tx1a(13) = -0.75;
    Global_Tx1a(14) = -0.6458333333333334;
    Global_Tx1a(15) = -0.5416666666666667;
    Global_Tx1a(16) = -0.4375000000000001;
    Global_Tx1a(17) = -0.3333333333333335;
    Global_Tx1a(18) = -0.22916666666666685;
    Global_Tx1a(19) = -0.12500000000000022;
    Global_Tx1a(20) = -0.020833333333333592;
    Global_Tx1a(21) = 0.08333333333333304;
    Global_Tx1a(22) = 0.18749999999999967;
    Global_Tx1a(23) = 0.2916666666666663;
    Global_Tx1a(24) = 0.3958333333333329;
    Global_Tx1a(25) = 0.5;
    Global_Tx1a(26) = 0.5520833333333333;
    Global_Tx1a(27) = 0.6041666666666666;
    Global_Tx1a(28) = 0.65625;
    Global_Tx1a(29) = 0.7083333333333333;
    Global_Tx1a(30) = 0.7604166666666665;
    Global_Tx1a(31) = 0.8124999999999999;
    Global_Tx1a(32) = 0.8645833333333331;
    Global_Tx1a(33) = 0.9166666666666665;
    Global_Tx1a(34) = 0.9687499999999998;
    Global_Tx1a(35) = 1.020833333333333;
    Global_Tx1a(36) = 1.0729166666666663;
    Global_Tx1a(37) = 1.1249999999999998;
    Global_Tx1a(38) = 1.177083333333333;
    Global_Tx1a(39) = 1.2291666666666665;
    Global_Tx1a(40) = 1.2812499999999998;
    Global_Tx1a(41) = 1.333333333333333;
    Global_Tx1a(42) = 1.3854166666666663;
    Global_Tx1a(43) = 1.4374999999999996;
    Global_Tx1a(44) = 1.489583333333333;
    Global_Tx1a(45) = 1.5416666666666663;
    Global_Tx1a(46) = 1.5937499999999996;
    Global_Tx1a(47) = 1.645833333333333;
    Global_Tx1a(48) = 1.6979166666666665;
    Global_Tx1a(49) = 1.7499999999999996;
    Global_Tx1a(50) = 1.802083333333333;
    Global_Tx1a(51) = 1.854166666666666;
    Global_Tx1a(52) = 1.9062499999999996;
    Global_Tx1a(53) = 1.958333333333333;
    Global_Tx1a(54) = 2.0104166666666665;
    Global_Tx1a(55) = 2.0624999999999996;
    Global_Tx1a(56) = 2.166666666666666;
    Global_Tx1a(57) = 2.2708333333333326;
    Global_Tx1a(58) = 2.3749999999999996;
    Global_Tx1a(59) = 2.479166666666666;
    Global_Tx1a(60) = 2.583333333333333;
    Global_Tx1a(61) = 2.6875;
    Global_Tx1a(62) = 2.7916666666666665;
    Global_Tx1a(63) = 2.895833333333333;
    Global_Tx1a(64) = 3.;

    Global_Tya.SetShape(64 );
    Global_Tya(1) = 7.090702573174318;
    Global_Tya(2) = 6.438211785191379;
    Global_Tya(3) = 5.817695631982345;
    Global_Tya(4) = 5.229458398114625;
    Global_Tya(5) = 4.673689698348237;
    Global_Tya(6) = 4.15046242204372;
    Global_Tya(7) = 3.6597319429768445;
    Global_Tya(8) = 3.201336603126446;
    Global_Tya(9) = 2.7749994651906897;
    Global_Tya(10) = 2.380331314832929;
    Global_Tya(11) = 2.0168348801101734;
    Global_Tya(12) = 1.6839102223417899;
    Global_Tya(13) = 1.380861239976666;
    Global_Tya(14) = 1.106903214951137;
    Global_Tya(15) = 0.8611713197284716;
    Global_Tya(16) = 0.6427299927960622;
    Global_Tya(17) = 0.4505830809816258;
    Global_Tya(18) = 0.28368463863670784;
    Global_Tya(19) = 0.14095026661477258;
    Global_Tya(20) = 0.0212688681193019;
    Global_Tya(21) = -0.07648530602191173;
    Global_Tya(22) = -0.15344045323772992;
    Global_Tya(23) = -0.21071498855336057;
    Global_Tya(24) = -0.24940541997997812;
    Global_Tya(25) = -0.270574461395797;
    Global_Tya(26) = -0.2749084736255334;
    Global_Tya(27) = -0.27523951178550565;
    Global_Tya(28) = -0.2716858604242085;
    Global_Tya(29) = -0.2643616257333199;
    Global_Tya(30) = -0.25337642608480904;
    Global_Tya(31) = -0.23883509473928766;
    Global_Tya(32) = -0.2208373955318681;
    Global_Tya(33) = -0.19947775230661358;
    Global_Tya(34) = -0.1748449928333633;
    Global_Tya(35) = -0.14702210790145265;
    Global_Tya(36) = -0.11608602624367403;
    Global_Tya(37) = -0.08210740590090504;
    Global_Tya(38) = -0.04515044259322876;
    Global_Tya(39) = -0.005272695617261736;
    Global_Tya(40) = 0.03747506825813973;
    Global_Tya(41) = 0.08304901247442364;
    Global_Tya(42) = 0.13141234957074088;
    Global_Tya(43) = 0.18253544095376117;
    Global_Tya(44) = 0.23639587727933176;
    Global_Tya(45) = 0.2929785392279691;
    Global_Tya(46) = 0.35227563850937527;
    Global_Tya(47) = 0.4142867389847429;
    Global_Tya(48) = 0.47901875784959635;
    Global_Tya(49) = 0.5464859468739363;
    Global_Tya(50) = 0.6167098537505806;
    Global_Tya(51) = 0.6897192636565137;
    Global_Tya(52) = 0.765550121185754;
    Global_Tya(53) = 0.844245432865435;
    Global_Tya(54) = 0.9258551505195092;
    Global_Tya(55) = 1.0104360357963778;
    Global_Tya(56) = 1.188771463237071;
    Global_Tya(57) = 1.3798357076052792;
    Global_Tya(58) = 1.5843100319532715;
    Global_Tya(59) = 1.8029655844723624;
    Global_Tya(60) = 2.036655038562893;
    Global_Tya(61) = 2.286303349098633;
    Global_Tya(62) = 2.5528977250959644;
    Global_Tya(63) = 2.837476927492947;
    Global_Tya(64) = 3.1411200080598665;
; 
}

double Global_TInterpolatingFunction1D ( const double &x)
{
    int n;
    int j;
    int m;
    int k;
    n = Global_Tx1a.dimension(1);
    j = Global_TLocate (Global_Tx1a, x);
    m = 4;
    k = Min(Max(j+-ifloor((m+-1.)/2),1),n+1+-m);
    return Global_TPolyInt (Global_Tx1a(R(k,k+m+-1)), \
Global_Tya(R(k,k+m+-1)), x
        );
}

int Global_TLocate ( const doubleN &xx, const double &x)
{
    int n;
    int ju;
    int jm;
    int jl;
    int j;
    n = xx.dimension(1);
    jl = 0;
    ju = n+1;
    while (ju+-jl > 1)
    {
        jm = ifloor((ju+jl)/2.);
        if (x >= xx(jm) == xx(n) > xx(1))
        {
            jl = jm;
        }
        else
        {
            ju = jm;
        }
    }
    return  (( x == xx(1) ) ? ( j = 1 ) : (  (( x == xx(n) ) ? ( j = n+-1 ) : \
( 
        j = jl ))  )) ;
}

double Global_TPolyInt ( const doubleN &xa, const doubleN &ya, const double &x)
{
    int nn = xa.dimension(1);
    int n;
    int i;
    int m;
    int ns = 1;
    double den;
    double dif;
    double dift;
    double ho;
    double hp;
    double w;
    doubleN c(nn);
    doubleN d(nn);
    double y;
    int i_D3067;
    int i_D3069;
    int mctmp_D3062_D3072;
    int mctmp_D3065_D3074;
    int m_D3076;


    n = xa.dimension(1);
    dif = abs(x+-xa(1));
    c.SetShape(Max(n,0) );
    mctmp_D3062_D3072 = 1;
    while (mctmp_D3062_D3072 <= n)
    {
        c(mctmp_D3062_D3072) = mctmp_D3062_D3072;
        mctmp_D3062_D3072 = mctmp_D3062_D3072+1;
    }
    d.SetShape(Max(n,0) );
    mctmp_D3065_D3074 = 1;
    while (mctmp_D3065_D3074 <= n)
    {
        d(mctmp_D3065_D3074) = mctmp_D3065_D3074;
        mctmp_D3065_D3074 = mctmp_D3065_D3074+1;
    }
    i_D3067 = 1;
    while (i_D3067 <= n)
    {
        dift = abs(x+-xa(i_D3067));
        if (dift < dif)
        {
            ns = i_D3067;
            dif = dift;
        }
        c(i_D3067) = ya(i_D3067);
        d(i_D3067) = ya(i_D3067);
        i_D3067 = i_D3067+1;
    }
    y = ya(ns);
    ns = ns+-1;
    m_D3076 = 1;
    while (m_D3076 <= n+-1)
    {
        i_D3069 = 1;
        while (i_D3069 <= n+-m_D3076)
        {
            ho = xa(i_D3069)+-x;
            hp = xa(i_D3069+m_D3076)+-x;
            w = c(i_D3069+1)+-d(i_D3069);
            den = ho+-hp;
            den = w/den;
            d(i_D3069) = hp*den;
            c(i_D3069) = ho*den;
            i_D3069 = i_D3069+1;
        }
        y = y+ (( 2*ns < n+-m_D3076 ) ? ( c(ns+1) ) : (  ( ns = ns+-1,d(ns+1) \
) 
             )) ;
        m_D3076 = m_D3076+1;
    }
    return y;
}

doubleN Global_Tya(64);
doubleN Global_Tx1a(64);\
\>", "Print",
 CellChangeTimes->{3.5086608389659452`*^9, 3.508660886777834*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Stand-alone executable generation", "Subsection"],

Cell[TextData[{
 "We generate a function which is ",
 StyleBox["Mathematica - ",
  FontSlant->"Italic"],
 "independent.",
 " The option ",
 StyleBox["MainFileAndFunction",
  FontFamily->"Courier New"],
 " a string which respresents  main.cc for stand-alone code is given."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"CompilePackage", "[", "\[IndentingNewLine]", 
  RowBox[{
  "MainFileAndFunction", "\[Rule]", "\n", "\t", 
   "\"\<\n\t\tint main(){\n\t\tGlobal_TGlobalInit();\n\t\tcout << \
Global_TInterpolatingFunction1D(-1.1);\n\t}\n\t\>\""}], "\[IndentingNewLine]",
   "]"}]], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "f", "}"}]}],
  SequenceForm["Functions without types are not compiled :", {$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660839247192*^9, 3.508660887074705*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "f", "}"}]}],
  SequenceForm["Functions without types are not compiled :", {$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660839247192*^9, 3.508660887121579*^9}],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Functions without types are not compiled :\"\>", 
   "\[InvisibleSpace]", 
   RowBox[{"{", "f", "}"}]}],
  SequenceForm["Functions without types are not compiled :", {$CellContext`f}],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660839247192*^9, 3.5086608871684537`*^9}],

Cell[BoxData["\<\"Successful compilation to C++: 3 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508660839247192*^9, 3.5086608893559256`*^9}]
}, Open  ]],

Cell[TextData[{
 "Option ",
 StyleBox["StandAloneExecutable",
  FontFamily->"Courier New"],
 " for  ",
 StyleBox["MakeBinary",
  FontFamily->"Courier"],
 " specifies that stand-alone executable will be linked."
}], "Text"],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", 
  RowBox[{"StandAloneExecutable", "\[Rule]", "True"}], "]"}]], "Input"],

Cell["Start Dos command line prompt", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"If", "[", 
   RowBox[{"$Notebooks", ",", 
    RowBox[{"Run", "[", "\"\<command\>\"", "]"}]}], "]"}], ";"}]], "Input"],

Cell[BoxData[
 RowBox[{"RunThrough", "[", 
  RowBox[{"\"\<Global.exe\>\"", ",", "\"\<\>\""}], "]"}]], "Input"],

Cell[TextData[{
 "Stand-alone program is ",
 StyleBox["global.exe",
  FontFamily->"Courier New"],
 ". It gives the same value as above. Close Dos window after use."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"InterpolatingFunction1D", "[", 
  RowBox[{"-", "1.1"}], "]"}]], "Input"],

Cell[BoxData["2.51879489779062`"], "Output",
 CellChangeTimes->{3.508660861371909*^9, 3.508660954573841*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"Directory", "[", "]"}]], "Input"],

Cell[BoxData["\<\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\Interpolation-1D\"\
\>"], "Output",
 CellChangeTimes->{3.5086608613875337`*^9, 3.5086609545894656`*^9}]
}, Open  ]]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Conclusion", "Section"],

Cell[TextData[{
 "Information in ",
 StyleBox["InterpolationgFunction",
  FontFamily->"Courier New"],
 " object extracted and saved in variables  x1 och y. The interpolation \
function is defined after that with help of help-functions. C++ code is \
automatically generated, compiled and tested using plotting the difference. \
Difference is very small. Creation of stand-alone function is illustrated.   \
"
}], "Text"],

Cell["Removal of files and definitions.", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"CleanMathCodeFiles", "[", 
   RowBox[{
    RowBox[{"CleanAllBut", "\[Rule]", 
     RowBox[{"{", "}"}]}], ",", 
    RowBox[{"Confirm", "\[Rule]", "False"}]}], "]"}], ";"}]], "Input"]
}, Open  ]]
},
WindowToolbars->"EditBar",
WindowSize->{998, 823},
WindowMargins->{{65, Automatic}, {19, Automatic}},
Magnification->1.25,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->"Classroom.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 36, 0, 69, "Title"],
Cell[619, 23, 74, 0, 46, "Subtitle"],
Cell[696, 25, 17, 0, 69, "Title"],
Cell[716, 27, 27, 3, 66, "Author"],
Cell[CellGroupData[{
Cell[768, 34, 31, 0, 67, "Section"],
Cell[802, 36, 550, 10, 104, "Text"],
Cell[1355, 48, 16, 0, 29, "Text"],
Cell[CellGroupData[{
Cell[1396, 52, 92, 2, 57, "Input"],
Cell[1491, 56, 545, 11, 27, "Print"]
}, Open  ]],
Cell[2051, 70, 122, 3, 57, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2210, 78, 44, 0, 67, "Section"],
Cell[2257, 80, 90, 2, 29, "Text"],
Cell[2350, 84, 187, 6, 59, "Input"],
Cell[2540, 92, 293, 14, 30, "Text"],
Cell[CellGroupData[{
Cell[2858, 110, 345, 9, 57, "Input"],
Cell[3206, 121, 394, 12, 57, "Output"]
}, Open  ]],
Cell[3615, 136, 467, 20, 31, "Text"],
Cell[4085, 158, 231, 6, 57, "Input"],
Cell[4319, 166, 939, 23, 190, "Input"],
Cell[CellGroupData[{
Cell[5283, 193, 62, 1, 80, "Input"],
Cell[5348, 196, 2411, 44, 190, "Output"]
}, Open  ]],
Cell[7774, 243, 117, 3, 29, "Text"],
Cell[7894, 248, 599, 17, 168, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[8530, 270, 60, 0, 67, "Section"],
Cell[8593, 272, 26, 0, 29, "Text"],
Cell[CellGroupData[{
Cell[8644, 276, 4681, 116, 674, "Input"],
Cell[13328, 394, 612, 12, 27, "Message"]
}, Open  ]],
Cell[13955, 409, 125, 3, 29, "Text"],
Cell[14083, 414, 2336, 60, 533, "Input"],
Cell[16422, 476, 60, 0, 29, "Text"],
Cell[16485, 478, 1332, 36, 227, "Input"],
Cell[17820, 516, 112, 3, 29, "Text"],
Cell[17935, 521, 668, 17, 80, "Input"],
Cell[18606, 540, 397, 12, 57, "Input"],
Cell[19006, 554, 296, 8, 57, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[19339, 567, 34, 0, 67, "Section"],
Cell[19376, 569, 116, 4, 29, "Text"],
Cell[CellGroupData[{
Cell[19517, 577, 62, 1, 57, "Input"],
Cell[19582, 580, 377, 8, 27, "Print"],
Cell[19962, 590, 380, 8, 27, "Print"],
Cell[20345, 600, 378, 8, 27, "Print"],
Cell[20726, 610, 195, 2, 27, "Print"]
}, Open  ]],
Cell[20936, 615, 174, 5, 29, "Text"],
Cell[21113, 622, 58, 1, 57, "Input"],
Cell[CellGroupData[{
Cell[21196, 627, 26, 0, 54, "Subsection"],
Cell[21225, 629, 231, 8, 29, "Text"],
Cell[21459, 639, 47, 0, 29, "Text"],
Cell[CellGroupData[{
Cell[21531, 643, 74, 2, 57, "Input"],
Cell[21608, 647, 162, 2, 76, "Output"]
}, Open  ]],
Cell[21785, 652, 112, 4, 29, "Text"],
Cell[CellGroupData[{
Cell[21922, 660, 74, 2, 57, "Input"],
Cell[21999, 664, 157, 2, 76, "Output"]
}, Open  ]],
Cell[22171, 669, 142, 3, 29, "Text"],
Cell[CellGroupData[{
Cell[22338, 676, 96, 2, 57, "Input"],
Cell[22437, 680, 161, 2, 76, "Output"]
}, Open  ]],
Cell[22613, 685, 108, 5, 29, "Text"],
Cell[CellGroupData[{
Cell[22746, 694, 77, 2, 57, "Input"],
Cell[22826, 698, 301, 6, 27, "Print"]
}, Open  ]],
Cell[CellGroupData[{
Cell[23164, 709, 96, 2, 57, "Input"],
Cell[23263, 713, 110, 1, 76, "Output"]
}, Open  ]],
Cell[23388, 717, 79, 2, 57, "Input"],
Cell[23470, 721, 182, 7, 29, "Text"],
Cell[CellGroupData[{
Cell[23677, 732, 344, 10, 57, "Input"],
Cell[24024, 744, 44649, 736, 346, "Output"]
}, Open  ]],
Cell[68688, 1483, 97, 2, 29, "Text"],
Cell[CellGroupData[{
Cell[68810, 1489, 38, 0, 57, "Input"],
Cell[68851, 1491, 103, 1, 76, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[68991, 1497, 108, 3, 57, "Input"],
Cell[69102, 1502, 94, 1, 76, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[69233, 1508, 431, 13, 80, "Input"],
Cell[69667, 1523, 23343, 398, 338, "Output"]
}, Open  ]],
Cell[93025, 1924, 29, 0, 29, "Text"],
Cell[93057, 1926, 43, 0, 29, "Text"],
Cell[CellGroupData[{
Cell[93125, 1930, 77, 2, 57, "Input"],
Cell[93205, 1934, 248, 5, 27, "Print"]
}, Open  ]],
Cell[93468, 1942, 133, 4, 30, "Text"],
Cell[CellGroupData[{
Cell[93626, 1950, 687, 21, 80, "Input"],
Cell[94316, 1973, 73739, 1218, 338, "Output"]
}, Open  ]],
Cell[168070, 3194, 30, 0, 29, "Text"],
Cell[168103, 3196, 30, 0, 29, "Text"],
Cell[168136, 3198, 119, 4, 80, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[168292, 3207, 30, 0, 54, "Subsection"],
Cell[CellGroupData[{
Cell[168347, 3211, 78, 1, 57, "Input"],
Cell[168428, 3214, 8056, 264, 4736, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[176533, 3484, 55, 0, 54, "Subsection"],
Cell[176591, 3486, 284, 9, 50, "Text"],
Cell[CellGroupData[{
Cell[176900, 3499, 296, 6, 234, "Input"],
Cell[177199, 3507, 327, 7, 27, "Print"],
Cell[177529, 3516, 327, 7, 27, "Print"],
Cell[177859, 3525, 329, 7, 27, "Print"],
Cell[178191, 3534, 144, 1, 27, "Print"]
}, Open  ]],
Cell[178350, 3538, 222, 8, 29, "Text"],
Cell[178575, 3548, 114, 2, 57, "Input"],
Cell[178692, 3552, 45, 0, 29, "Text"],
Cell[178740, 3554, 152, 4, 57, "Input"],
Cell[178895, 3560, 110, 2, 57, "Input"],
Cell[179008, 3564, 176, 5, 29, "Text"],
Cell[CellGroupData[{
Cell[179209, 3573, 96, 2, 57, "Input"],
Cell[179308, 3577, 108, 1, 76, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[179453, 3583, 57, 1, 57, "Input"],
Cell[179513, 3586, 162, 2, 76, "Output"]
}, Open  ]]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[179736, 3595, 29, 0, 67, "Section"],
Cell[179768, 3597, 420, 9, 71, "Text"],
Cell[180191, 3608, 49, 0, 29, "Text"],
Cell[180243, 3610, 216, 6, 57, "Input"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
